from __future__ import unicode_literals

import cStringIO
import logging

from celery import shared_task
from django.conf import settings
from django.core.mail import send_mail
from django.db.models import Q
from django.template.loader import render_to_string
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_core.core.csv import UnicodeDictWriter
from waldur_core.structure import models as structure_models
from waldur_mastermind.invoices.utils import get_previous_month

from . import models, registrators, serializers, utils


logger = logging.getLogger(__name__)


@shared_task(name='invoices.create_monthly_invoices')
def create_monthly_invoices():
    """
    - For every customer change state of the invoices for previous months from "pending" to "billed"
      and freeze their items.
    - Create new invoice for every customer in current month if not created yet.
    """
    date = timezone.now()

    old_invoices = models.Invoice.objects.filter(
        Q(state=models.Invoice.States.PENDING, year__lt=date.year) |
        Q(state=models.Invoice.States.PENDING, year=date.year, month__lt=date.month)
    )
    for invoice in old_invoices:
        invoice.set_created()
        invoice.freeze()

    customers = structure_models.Customer.objects.all()
    if settings.WALDUR_CORE['ENABLE_ACCOUNTING_START_DATE']:
        customers = customers.filter(accounting_start_date__lt=timezone.now())

    for customer in customers.iterator():
        registrators.RegistrationManager.get_or_create_invoice(customer, core_utils.month_start(date))

    if settings.WALDUR_INVOICES['INVOICE_REPORTING']['ENABLE']:
        send_invoice_report.delay()


@shared_task(name='invoices.send_invoice_notification')
def send_invoice_notification(invoice_uuid, link_template):
    """ Sends email notification with invoice link to customer owners """
    invoice = models.Invoice.objects.get(uuid=invoice_uuid)

    context = {
        'month': invoice.month,
        'year': invoice.year,
        'customer': invoice.customer.name,
        'link': link_template.format(uuid=invoice_uuid)
    }

    subject = render_to_string('invoices/notification_subject.txt', context).strip()
    text_message = render_to_string('invoices/notification_message.txt', context)
    html_message = render_to_string('invoices/notification_message.html', context)

    emails = [owner.email for owner in invoice.customer.get_owners()]

    logger.debug('About to send invoice {invoice} notification to {emails}'.format(invoice=invoice, emails=emails))
    send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, emails, html_message=html_message)


@shared_task(name='invoices.send_invoice_report')
def send_invoice_report():
    """ Sends aggregate accounting data as CSV """
    date = get_previous_month()
    subject = render_to_string('invoices/report_subject.txt', {
        'month': date.month,
        'year': date.year,
    }).strip()
    body = render_to_string('invoices/report_body.txt', {
        'month': date.month,
        'year': date.year,
    }).strip()
    filename = '3M%02d%dWaldur.txt' % (date.month, date.year)
    invoices = models.Invoice.objects.filter(year=date.year, month=date.month)

    # Report should include only organizations that had accounting running during the invoice period.
    if settings.WALDUR_CORE['ENABLE_ACCOUNTING_START_DATE']:
        invoices = invoices.filter(customer__accounting_start_date__lte=date)

    # Report should not include customers with 0 invoice sum.
    invoices = [invoice for invoice in invoices if invoice.total > 0]
    text_message = format_invoice_csv(invoices)

    # Please note that email body could be empty if there are no valid invoices
    emails = [settings.WALDUR_INVOICES['INVOICE_REPORTING']['EMAIL']]
    logger.debug('About to send accounting report to {emails}'.format(emails=emails))
    utils.send_mail_attachment(
        subject=subject,
        body=body,
        to=emails,
        attach_text=text_message,
        filename=filename
    )


def format_invoice_csv(invoices):
    if not isinstance(invoices, list):
        invoices = [invoices]

    csv_params = settings.WALDUR_INVOICES['INVOICE_REPORTING']['CSV_PARAMS']

    if settings.WALDUR_INVOICES['INVOICE_REPORTING'].get('USE_SAF'):
        fields = serializers.SAFReportSerializer.Meta.fields
        stream = cStringIO.StringIO()
        writer = UnicodeDictWriter(stream, fieldnames=fields, **csv_params)
        writer.writeheader()

        for invoice in invoices:
            serializer = serializers.SAFReportSerializer(invoice.items, many=True)
            writer.writerows(serializer.data)
        return stream.getvalue().decode('utf-8')

    fields = serializers.InvoiceItemReportSerializer.Meta.fields
    stream = cStringIO.StringIO()
    writer = UnicodeDictWriter(stream, fieldnames=fields, **csv_params)
    writer.writeheader()

    for invoice in invoices:
        openstack_items = invoice.openstack_items.all().select_related('invoice__customer')
        openstack_serializer = serializers.OpenStackItemReportSerializer(openstack_items, many=True)
        writer.writerows(openstack_serializer.data)

        offering_items = invoice.offering_items.all().select_related('invoice__customer')
        offering_serializer = serializers.OfferingItemReportSerializer(offering_items, many=True)
        writer.writerows(offering_serializer.data)

        generic_items = invoice.generic_items.all().select_related('invoice__customer')
        generic_serializer = serializers.GenericItemReportSerializer(generic_items, many=True)
        writer.writerows(generic_serializer.data)

    return stream.getvalue().decode('utf-8')


@shared_task(name='invoices.update_invoices_current_cost')
def update_invoices_current_cost():
    year = utils.get_current_year()
    month = utils.get_current_month()

    for invoice in models.Invoice.objects.filter(year=year, month=month):
        invoice.update_current_cost()



# Django settings for generating documentation for Waldur Core.
from waldur_core.server.base_settings import *  # noqa

SECRET_KEY = 'test-key'

DEBUG = True

ALLOWED_HOSTS = ['127.0.0.1']

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake',
    }
}

CELERY_BROKER_URL = 'sqla+sqlite:///:memory:'
CELERY_RESULT_BACKEND = 'db+sqlite:///:memory:'



import logging

from waldur_mastermind.invoices import registrators
from waldur_mastermind.invoices import models as invoice_models

from . import models

logger = logging.getLogger(__name__)


class ExpertRequestRegistrator(registrators.BaseRegistrator):
    def get_sources(self, customer):
        return models.ExpertRequest.objects.filter(project__customer=customer,
                                                   state=models.ExpertRequest.States.COMPLETED)

    def get_customer(self, source):
        return source.customer

    def _find_item(self, source, now):
        expert_request = source
        result = invoice_models.GenericInvoiceItem.objects.filter(
            scope=expert_request,
            invoice__customer=self.get_customer(expert_request),
            invoice__state=invoice_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        ).first()

        return result

    def _create_item(self, source, invoice, start, end):
        expert_request = source

        if (not expert_request.recurring_billing and
                invoice_models.GenericInvoiceItem.objects.filter(scope=expert_request).exists()):
            return

        return invoice_models.GenericInvoiceItem.objects.create(
            scope=expert_request,
            project=expert_request.project,
            unit_price=self.get_price(expert_request),
            unit=invoice_models.GenericInvoiceItem.Units.QUANTITY,
            quantity=1,
            product_code=expert_request.product_code,
            article_code=expert_request.article_code,
            invoice=invoice,
            start=start,
            end=end,
        )

    def get_price(self, expert_request):
        return expert_request.contract.price

    def get_details(self, source):
        expert_request = source

        return {
            'expert_request_type': expert_request.type,
            'expert_request_recurring': expert_request.recurring_billing,
            'expert_request_name': expert_request.name,
            'issue_summary': expert_request.issue.summary,
            'issue_key': expert_request.issue.key,
            'issue_link': expert_request.issue.link,
        }



from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, exceptions, viewsets, response, status, serializers as rf_serializers

from waldur_core.core import validators as core_validators
from waldur_core.structure import views as structure_views

from . import models, serializers, executors, filters
from .backend import SizeQueryset


class AzureServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.AzureService.objects.all()
    serializer_class = serializers.ServiceSerializer
    import_serializer_class = serializers.VirtualMachineImportSerializer


class AzureServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.AzureServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    filter_class = filters.ImageFilter
    lookup_field = 'uuid'

    def get_queryset(self):
        return models.Image.objects.order_by('name')


class SizeViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = SizeQueryset()
    serializer_class = serializers.SizeSerializer
    lookup_field = 'uuid'


class VirtualMachineViewSet(structure_views.BaseResourceViewSet):
    queryset = models.VirtualMachine.objects.all()
    serializer_class = serializers.VirtualMachineSerializer
    delete_executor = executors.VirtualMachineDeleteExecutor

    @decorators.detail_route()
    def rdp(self, request, uuid=None):
        vm = self.get_object()

        try:
            rdp_endpoint = vm.endpoints.get(name=models.InstanceEndpoint.Name.RDP)
        except models.InstanceEndpoint.DoesNotExist:
            raise exceptions.NotFound("This virtual machine doesn't run remote desktop")

        response = HttpResponse(content_type='application/x-rdp')
        response['Content-Disposition'] = 'attachment; filename="{}.rdp"'.format(vm.name)
        response.write(
            "full address:s:%s.cloudapp.net:%s\n"
            "prompt for credentials:i:1\n\n" % (vm.service_project_link.cloud_service_name, rdp_endpoint.public_port))

        return response

    rdp_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                      core_validators.RuntimeStateValidator('running')]

    @decorators.detail_route(methods=['post'])
    def start(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineStartExecutor().execute(virtual_machine)
        return response.Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    start_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                        core_validators.RuntimeStateValidator('stopped')]
    start_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def stop(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineStopExecutor().execute(virtual_machine)
        return response.Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    stop_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                       core_validators.RuntimeStateValidator('running')]
    stop_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def restart(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineRestartExecutor().execute(virtual_machine)
        return response.Response({'status': _('restart was scheduled')}, status=status.HTTP_202_ACCEPTED)

    restart_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                          core_validators.RuntimeStateValidator('running')]
    restart_serializer_class = rf_serializers.Serializer

    def perform_create(self, serializer):
        instance = serializer.save()
        executors.VirtualMachineCreateExecutor.execute(
            instance,
            backend_image_id=serializer.validated_data['image'].backend_id,
            backend_size_id=serializer.validated_data['size'].pk,
        )



import json
from zipfile import is_zipfile, ZipFile

from django import forms
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from jsoneditor.forms import JSONEditor

from . import models


class ChangePlaybookParameterInline(admin.TabularInline):
    model = models.PlaybookParameter
    extra = 0
    readonly_fields = ('name', 'description', 'required', 'default')

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


class AddPlaybookParameterInline(admin.TabularInline):
    model = models.PlaybookParameter
    extra = 3
    fields = ('name', 'description', 'required', 'default')


class ChangePlaybookAdminForm(forms.ModelForm):
    class Meta:
        model = models.Playbook
        fields = ('name', 'description', 'image', 'entrypoint',)


class AddPlaybookAdminForm(forms.ModelForm):
    archive = forms.FileField()

    class Meta:
        model = models.Playbook
        fields = ('name', 'description', 'image', 'entrypoint')

    def clean_archive(self):
        value = self.cleaned_data['archive']
        if not is_zipfile(value):
            raise ValidationError(_('ZIP file must be uploaded.'))
        elif not value.name.endswith('.zip'):
            raise ValidationError(_("File must have '.zip' extension."))

        zip_file = ZipFile(value)
        invalid_file = zip_file.testzip()
        if invalid_file is not None:
            raise ValidationError(
                _('File {filename} in archive {archive_name} has an invalid type.'.format(
                    filename=invalid_file, archive_name=zip_file.filename)))

        return value

    def clean(self):
        cleaned_data = super(AddPlaybookAdminForm, self).clean()
        if self.errors:
            return cleaned_data

        archive = cleaned_data['archive']
        entrypoint = cleaned_data['entrypoint']

        zip_file = ZipFile(archive)
        if entrypoint not in zip_file.namelist():
            raise ValidationError(
                _('Failed to find entrypoint {entrypoint} in archive {archive_name}.'.format(
                    entrypoint=entrypoint, archive_name=zip_file.filename)))

        return cleaned_data

    def save(self, commit=True):
        self.instance.workspace = models.Playbook.generate_workspace_path()
        archive = self.cleaned_data.pop('archive')
        zip_file = ZipFile(archive)
        zip_file.extractall(self.instance.workspace)
        zip_file.close()

        return super(AddPlaybookAdminForm, self).save(commit)


class PlaybookAdmin(admin.ModelAdmin):
    list_filter = ('name', 'description')
    list_display = ('name', 'description')

    def get_readonly_fields(self, request, obj=None):
        if obj is not None:
            return self.readonly_fields + ('entrypoint',)
        return self.readonly_fields

    def add_view(self, request, form_url='', extra_context=None):
        self.inlines = [AddPlaybookParameterInline]
        self.form = AddPlaybookAdminForm
        return super(PlaybookAdmin, self).add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        self.inlines = [ChangePlaybookParameterInline]
        self.form = ChangePlaybookAdminForm
        return super(PlaybookAdmin, self).change_view(request, object_id, form_url, extra_context)


class JobAdminForm(forms.ModelForm):
    class Meta:
        widgets = {
            'arguments': JSONEditor(),
        }

    def clean(self):
        cleaned_data = super(JobAdminForm, self).clean()
        if self.errors:
            return cleaned_data

        if self.instance and self.instance.pk:
            playbook = self.instance.playbook
        else:
            playbook = cleaned_data['playbook']

        arguments = json.loads(cleaned_data.get('arguments'))
        parameter_names = playbook.parameters.all().values_list('name', flat=True)
        for argument in arguments.keys():
            if argument not in parameter_names and argument != 'project_uuid':
                raise ValidationError(_('Argument %s is not listed in playbook parameters.' % argument))

        if playbook.parameters.exclude(name__in=arguments.keys()).filter(required=True, default__exact='').exists():
            raise ValidationError(_('Not all required playbook parameters were specified.'))

        return cleaned_data

    def save(self, commit=True):
        arguments = self.instance.arguments
        playbook = self.instance.playbook
        unfilled_parameters = playbook.parameters.exclude(name__in=arguments.keys())
        for parameter in unfilled_parameters:
            if parameter.default:
                arguments[parameter.name] = parameter.default
        return super(JobAdminForm, self).save(commit)


class JobAdmin(admin.ModelAdmin):
    form = JobAdminForm
    fields = ('name', 'description', 'state', 'service_project_link',
              'playbook', 'arguments', 'output')
    list_filter = ('name', 'description', 'service_project_link', 'playbook')
    list_display = ('name', 'state', 'service_project_link', 'playbook')
    readonly_fields = ('output', 'created', 'modified')

    def get_readonly_fields(self, request, obj=None):
        if obj is not None:
            return self.readonly_fields + ('playbook',)
        return self.readonly_fields


admin.site.register(models.Playbook, PlaybookAdmin)
admin.site.register(models.Job, JobAdmin)



from celery import chain

from waldur_core.core import executors as core_executors, tasks as core_tasks


class InvoiceCreateExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, invoice, serialized_invoice, **kwargs):
        tasks = [
            core_tasks.BackendMethodTask().si(serialized_invoice, 'create_invoice'),
            core_tasks.BackendMethodTask().si(serialized_invoice, 'download_invoice_pdf'),
            core_tasks.BackendMethodTask().si(serialized_invoice, 'send_invoice'),
            core_tasks.BackendMethodTask().si(serialized_invoice, 'pull_invoice'),
        ]

        return chain(*tasks)


class InvoicePullExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, invoice, serialized_invoice, **kwargs):
        tasks = [
            core_tasks.BackendMethodTask().si(serialized_invoice, 'pull_invoice'),
            core_tasks.BackendMethodTask().si(serialized_invoice, 'download_invoice_pdf'),
        ]

        return chain(*tasks)



""" This module calculates the cheapest price for deployment plans. """
import collections

from waldur_core.structure import models as structure_models

from . import register


def get_filtered_services(deployment_plan):
    """ Get services that fits deployment plan requirements """
    service_models = structure_models.Service.get_all_models()
    deployment_plan_certifications = deployment_plan.get_required_certifications()
    for model in service_models:
        services = (
            model.objects
            .filter(projects=deployment_plan.project)
            .select_related('settings')
            .prefetch_related('settings__certifications')
        )
        for service in services:
            if set(service.settings.certifications.all()).issuperset(deployment_plan_certifications):
                yield service


# http://stackoverflow.com/questions/11351032/named-tuple-and-optional-keyword-arguments
def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    T._defaults = tuple(prototype)  # added for easier access
    return T


# Abstract object that represents the best choice for a particular service.
OptimizedService = namedtuple_with_defaults(
    'OptimizedService', ('service', 'price', 'error_message'), {'error_message': ''})


class Strategy(object):
    """ Abstract. Defines how get the cheapest services setups for deployment plan. """

    def __init__(self, deployment_plan):
        self.deployment_plan = deployment_plan

    def get_optimized(self):
        """ Return list of OptimizedService objects """
        raise NotImplementedError()


class SingleServiceStrategy(Strategy):
    """ Optimize deployment plan for each service separately and return list
        of all available variants.
    """

    def _get_optimized_service(self, service):
        optimizer_class = register.Register.get_optimizer(service.settings.type)
        if optimizer_class:
            optimizer = optimizer_class()
            try:
                return optimizer.optimize(self.deployment_plan, service)
            except OptimizationError as e:
                return OptimizedService(service=service, price=None, error_message=str(e))

    def get_optimized(self):
        optimized = []
        for service in get_filtered_services(self.deployment_plan):
            optimized_service = self._get_optimized_service(service)
            if optimized_service:
                optimized.append(optimized_service)
        return optimized


# Optimizer should raise this error if it is impossible to setup
# deployment plan for service
class OptimizationError(Exception):
    pass


class Optimizer(object):
    """ Abstract. Descendant should define how to get the cheapest setup for a
        particular service.
    """

    def optimize(self, deployment_plan, service):
        """ Return the cheapest setup as OptimizedService object """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0009_tenant_service_verbose_name'),
    ]

    operations = [
        migrations.RenameField(
            model_name='floatingip',
            old_name='status',
            new_name='runtime_state',
        ),
    ]



import django_filters
from django_filters.widgets import BooleanWidget

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class OpenStackTenantServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='openstacktenant-detail', name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.OpenStackTenantServiceProjectLink


class FlavorFilter(structure_filters.ServicePropertySettingsFilter):

    o = django_filters.OrderingFilter(fields=('cores', 'ram', 'disk'))

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Flavor
        fields = dict({
            'cores': ['exact', 'gte', 'lte'],
            'ram': ['exact', 'gte', 'lte'],
            'disk': ['exact', 'gte', 'lte'],
        }, **{field: ['exact'] for field in structure_filters.ServicePropertySettingsFilter.Meta.fields})


class NetworkFilter(structure_filters.ServicePropertySettingsFilter):

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Network
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('type', 'is_external')


class SubNetFilter(structure_filters.ServicePropertySettingsFilter):
    network = core_filters.URLFilter(view_name='openstacktenant-network-detail', name='network__uuid')
    network_uuid = django_filters.UUIDFilter(name='network__uuid')

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.SubNet
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('ip_version', 'enable_dhcp')


class FloatingIPFilter(structure_filters.ServicePropertySettingsFilter):
    free = django_filters.BooleanFilter(name='internal_ip', lookup_expr='isnull', widget=BooleanWidget)

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.FloatingIP
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('runtime_state', 'address', 'is_booked')


class VolumeFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(name='instance__uuid')

    snapshot = core_filters.URLFilter(
        view_name='openstacktenant-snapshot-detail', name='restoration__snapshot__uuid')
    snapshot_uuid = django_filters.UUIDFilter(name='restoration__snapshot__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Volume
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('instance__name', 'instance_name'),
        ('size', 'size'),
    )


class SnapshotFilter(structure_filters.BaseResourceFilter):
    source_volume_uuid = django_filters.UUIDFilter(name='source_volume__uuid')
    source_volume = core_filters.URLFilter(view_name='openstacktenant-volume-detail', name='source_volume__uuid')
    backup_uuid = django_filters.UUIDFilter(name='backups__uuid')
    backup = core_filters.URLFilter(view_name='openstacktenant-backup-detail', name='backups__uuid')

    snapshot_schedule = core_filters.URLFilter(
        view_name='openstacktenant-snapshot-schedule-detail', name='snapshot_schedule__uuid')
    snapshot_schedule_uuid = django_filters.UUIDFilter(name='snapshot_schedule__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Snapshot
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('source_volume__name', 'source_volume_name'),
        ('size', 'size'),
    )


class InstanceFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Instance
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('internal_ips_set__ip4_address', 'internal_ips'),
        ('internal_ips_set__floating_ips__address', 'external_ips'),
    )


class BackupFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(name='instance__uuid')
    backup_schedule = core_filters.URLFilter(
        view_name='openstacktenant-backup-schedule-detail', name='backup_schedule__uuid')
    backup_schedule_uuid = django_filters.UUIDFilter(name='backup_schedule__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Backup


class BackupScheduleFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(name='instance__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.BackupSchedule


class SnapshotScheduleFilter(structure_filters.BaseResourceFilter):
    source_volume = core_filters.URLFilter(view_name='openstacktenant-volume-detail', name='source_volume__uuid')
    source_volume_uuid = django_filters.UUIDFilter(name='source_volume__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.SnapshotSchedule


class SecurityGroupFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.SecurityGroup


class ImageFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Image



from waldur_core.core import WaldurExtension


class FreeIPAExtension(WaldurExtension):
    class Settings:
        WALDUR_FREEIPA = {
            'HOSTNAME': 'ipa.example.com',
            'USERNAME': 'admin',
            'PASSWORD': 'secret',
            'VERIFY_SSL': True,
            'USERNAME_PREFIX': 'waldur_',
            'GROUPNAME_PREFIX': 'waldur_',
            'BLACKLISTED_USERNAMES': ['root'],
        }

    @staticmethod
    def django_app():
        return 'waldur_freeipa'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'waldur-freeipa-sync-groups': {
                'task': 'waldur_freeipa.sync_groups',
                'schedule': timedelta(minutes=10),
                'args': (),
            },
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-21 09:26
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0014_unique_backend_id_for_attachment'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='report',
            field=waldur_core.core.fields.JSONField(blank=True),
        ),
    ]



import logging

from waldur_core.structure.models import NewResource as Resource

from . import executors
from .models import Host


logger = logging.getLogger(__name__)


def delete_hosts_on_scope_deletion(sender, instance, name, source, target, **kwargs):
    if target != Resource.States.DELETING:
        return
    for host in Host.objects.filter(scope=instance):
        if host.state == Host.States.OK:
            executors.HostDeleteExecutor.execute(host)
        elif host.state == Host.States.ERRED:
            executors.HostDeleteExecutor.execute(host, force=True)
        else:
            logger.exception(
                'Instance %s host was in state %s on instance deletion.', instance, host.human_readable_state)
            host.set_erred()
            host.save()
            executors.HostDeleteExecutor.execute(host, force=True)


def refresh_database_connection(sender, instance, created=False, **kwargs):
    if not created and instance.type == 'Zabbix' and instance.tracker.has_changed('options'):
        instance.get_backend()._get_db_connection(force=True)



from . import views


def register_in(router):
    router.register(r'openstack', views.OpenStackServiceViewSet, base_name='openstack')
    router.register(r'openstack-images', views.ImageViewSet, base_name='openstack-image')
    router.register(r'openstack-flavors', views.FlavorViewSet, base_name='openstack-flavor')
    router.register(r'openstack-tenants', views.TenantViewSet, base_name='openstack-tenant')
    router.register(r'openstack-service-project-link', views.OpenStackServiceProjectLinkViewSet, base_name='openstack-spl')
    router.register(r'openstack-security-groups', views.SecurityGroupViewSet, base_name='openstack-sgp')
    router.register(r'openstack-floating-ips', views.FloatingIPViewSet, base_name='openstack-fip')
    router.register(r'openstack-networks', views.NetworkViewSet, base_name='openstack-network')
    router.register(r'openstack-subnets', views.SubNetViewSet, base_name='openstack-subnet')



import json

from django import forms
from django.contrib import admin
from django.contrib.auth.models import Group
from django.forms import ModelForm
from jsoneditor.forms import JSONEditor
import six

from waldur_core.logging import models
from waldur_core.logging.loggers import get_valid_events, get_event_groups


class JSONMultipleChoiceField(forms.MultipleChoiceField):

    def prepare_value(self, value):
        if isinstance(value, six.string_types):
            return json.loads(value)
        return value


class BaseHookForm(forms.ModelForm):
    event_types = JSONMultipleChoiceField(
        choices=lambda: [(e, e) for e in get_valid_events()],
        widget=forms.SelectMultiple(attrs={'size': '30'}),
        required=False,
    )

    event_groups = JSONMultipleChoiceField(
        choices=lambda: [(g, g) for g in get_event_groups()],
        widget=forms.SelectMultiple(attrs={'size': '30'}),
        required=False,
    )


class SystemNotificationForm(BaseHookForm):
    roles = JSONMultipleChoiceField(
        choices=[(g, g) for g in models.SystemNotification.get_valid_roles()],
        widget=forms.SelectMultiple(attrs={'size': '30'}),
        required=True,
    )

    class Meta:
        model = models.SystemNotification
        exclude = 'uuid',

    def __init__(self, *args, **kwargs):
        super(SystemNotificationForm, self).__init__(*args, **kwargs)
        self.fields['hook_content_type'].queryset = models.BaseHook.get_all_content_types()


class AlertAdminForm(ModelForm):
    class Meta:
        widgets = {
            'context': JSONEditor(),
        }


class SystemNotificationAdmin(admin.ModelAdmin):
    form = SystemNotificationForm
    list_display = ('name', 'hook_content_type', 'roles')


class AlertAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'alert_type', 'created', 'closed', 'scope', 'severity')
    list_filter = ('alert_type', 'created', 'closed', 'severity')
    ordering = ('alert_type',)
    base_model = models.Alert
    form = AlertAdminForm


class BaseHookAdmin(admin.ModelAdmin):
    form = BaseHookForm
    list_display = ('uuid', 'user', 'is_active', 'event_types', 'event_groups')


class WebHookAdmin(BaseHookAdmin):
    list_display = BaseHookAdmin.list_display + ('destination_url',)


class EmailHookAdmin(BaseHookAdmin):
    list_display = BaseHookAdmin.list_display + ('email',)


class PushHookAdmin(BaseHookAdmin):
    list_display = BaseHookAdmin.list_display + ('type', 'device_id')


# This hack is needed because core admin is imported several times.
if admin.site.is_registered(Group):
    admin.site.unregister(Group)

admin.site.register(models.Alert, AlertAdmin)
admin.site.register(models.SystemNotification, SystemNotificationAdmin)
admin.site.register(models.WebHook, WebHookAdmin)
admin.site.register(models.EmailHook, EmailHookAdmin)
admin.site.register(models.PushHook, PushHookAdmin)



from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status
from rest_framework.decorators import detail_route
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.response import Response

from waldur_core.core.views import ProtectedViewSet
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import models as structure_models
from waldur_core.users import models, filters, serializers, tasks


class InvitationViewSet(ProtectedViewSet):
    queryset = models.Invitation.objects.all().order_by('-created')
    serializer_class = serializers.InvitationSerializer
    filter_backends = (
        structure_filters.GenericRoleFilter,
        DjangoFilterBackend,
        filters.InvitationCustomerFilterBackend,
    )
    filter_class = filters.InvitationFilter
    lookup_field = 'uuid'

    def can_manage_invitation_with(self, customer, customer_role=None, project_role=None):
        user = self.request.user
        if user.is_staff:
            return True

        is_owner = customer.has_user(user, structure_models.CustomerRole.OWNER)
        can_manage_owners = settings.WALDUR_CORE['OWNERS_CAN_MANAGE_OWNERS']

        # It is assumed that either customer_role or project_role is not None
        if customer_role:
            return is_owner and can_manage_owners
        if project_role:
            return is_owner

    def perform_create(self, serializer):
        project = serializer.validated_data.get('project')
        if project:
            customer = project.customer
        else:
            customer = serializer.validated_data.get('customer')

        customer_role = serializer.validated_data.get('customer_role')
        project_role = serializer.validated_data.get('project_role')

        if not self.can_manage_invitation_with(customer, customer_role, project_role):
            raise PermissionDenied()

        invitation = serializer.save()
        tasks.send_invitation.delay(invitation.uuid.hex, self.request.user.full_name or self.request.user.username)

    @detail_route(methods=['post'])
    def send(self, request, uuid=None):
        invitation = self.get_object()

        if not self.can_manage_invitation_with(invitation.customer,
                                               invitation.customer_role,
                                               invitation.project_role):
            raise PermissionDenied()
        elif invitation.state == models.Invitation.State.ACCEPTED or \
                invitation.state == models.Invitation.State.CANCELED:
            raise ValidationError(_('Only pending and expired invitations can be resent.'))

        if invitation.state == models.Invitation.State.EXPIRED:
            invitation.state = models.Invitation.State.PENDING
            invitation.created = timezone.now()
            invitation.save()

        tasks.send_invitation.delay(invitation.uuid.hex, self.request.user.full_name or self.request.user.username)
        return Response({'detail': _('Invitation sending has been successfully scheduled.')},
                        status=status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def cancel(self, request, uuid=None):
        invitation = self.get_object()

        if not self.can_manage_invitation_with(invitation.customer,
                                               invitation.customer_role,
                                               invitation.project_role):
            raise PermissionDenied()
        elif invitation.state != models.Invitation.State.PENDING:
            raise ValidationError(_('Only pending invitation can be canceled.'))

        invitation.cancel()
        return Response({'detail': _('Invitation has been successfully canceled.')},
                        status=status.HTTP_200_OK)

    @detail_route(methods=['post'], filter_backends=[])
    def accept(self, request, uuid=None):
        """ Accept invitation for current user.

            To replace user's email with email from invitation - add parameter
            'replace_email' to request POST body.
        """
        invitation = self.get_object()

        if invitation.state != models.Invitation.State.PENDING:
            raise ValidationError(_('Only pending invitation can be accepted.'))
        elif invitation.civil_number and invitation.civil_number != request.user.civil_number:
            raise ValidationError(_('User has an invalid civil number.'))

        if invitation.project:
            if invitation.project.has_user(request.user):
                raise ValidationError(_('User already has role within this project.'))
        elif invitation.customer.has_user(request.user):
            raise ValidationError(_('User already has role within this customer.'))

        if settings.WALDUR_CORE['VALIDATE_INVITATION_EMAIL'] and invitation.email != request.user.email:
            raise ValidationError(_('Invitation and user emails mismatch.'))

        replace_email = bool(request.data.get('replace_email'))
        invitation.accept(request.user, replace_email=replace_email)
        return Response({'detail': _('Invitation has been successfully accepted.')},
                        status=status.HTTP_200_OK)

    @detail_route(methods=['post'], filter_backends=[], permission_classes=[])
    def check(self, request, uuid=None):
        invitation = self.get_object()

        if invitation.state != models.Invitation.State.PENDING:
            return Response(status=status.HTTP_404_NOT_FOUND)
        elif invitation.civil_number:
            return Response({'email': invitation.email, 'civil_number_required': True}, status=status.HTTP_200_OK)
        else:
            return Response({'email': invitation.email}, status=status.HTTP_200_OK)



from __future__ import unicode_literals

import base64
from decimal import Decimal
from HTMLParser import HTMLParser
import StringIO

from django.conf import settings
from django.core.validators import MinValueValidator
from django.db import models
from django.template.loader import render_to_string
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
import pdfkit
import six

from waldur_core.core import fields as core_fields
from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models
from waldur_mastermind.common import mixins as common_mixins, utils as common_utils
from waldur_mastermind.support import models as support_models

from . import managers


@python_2_unicode_compatible
class ExpertProvider(core_models.UuidMixin,
                     structure_models.TimeStampedModel):
    customer = models.OneToOneField(structure_models.Customer, related_name='+', on_delete=models.CASCADE)
    enable_notifications = models.BooleanField(default=True)

    class Meta(object):
        verbose_name = _('Expert providers')

    def __str__(self):
        return six.text_type(self.customer)

    @classmethod
    def get_url_name(cls):
        return 'expert-provider'


class PriceMixin(models.Model):
    class Meta(object):
        abstract = True

    price = models.DecimalField(max_digits=22, decimal_places=7,
                                validators=[MinValueValidator(Decimal('0'))],
                                default=0)


@python_2_unicode_compatible
class ExpertRequest(core_models.UuidMixin,
                    core_models.NameMixin,
                    PriceMixin,
                    common_mixins.ProductCodeMixin,
                    structure_models.StructureLoggableMixin,
                    structure_models.TimeStampedModel):
    class States(object):
        PENDING = 'pending'
        ACTIVE = 'active'
        CANCELLED = 'cancelled'
        COMPLETED = 'completed'

        CHOICES = (
            (PENDING, _('Pending')),
            (ACTIVE, _('Active')),
            (CANCELLED, _('Cancelled')),
            (COMPLETED, _('Completed'))
        )

    description = models.TextField(blank=True)
    user = models.ForeignKey(core_models.User, related_name='+', on_delete=models.CASCADE,
                             help_text=_('The user which has created this request.'))
    project = models.ForeignKey(structure_models.Project, related_name='+', on_delete=models.SET_NULL, null=True)
    # Project name, project UUID, customer should be stored separately
    # because they are not available after project removal
    project_name = models.CharField(max_length=150, blank=True)
    project_uuid = models.CharField(max_length=32, blank=True)
    customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE, null=True)

    state = models.CharField(default=States.PENDING, max_length=30, choices=States.CHOICES)
    type = models.CharField(max_length=255)
    extra = core_fields.JSONField(default=dict)
    issue = models.ForeignKey(support_models.Issue, null=True, on_delete=models.SET_NULL)
    recurring_billing = models.BooleanField(
        default=False, help_text=_('Defines whether expert request has to be billed every month or only once'))
    objectives = models.TextField(blank=True)
    milestones = models.TextField(blank=True)
    contract_methodology = models.TextField(blank=True)
    out_of_scope = models.TextField(
        blank=True, help_text=_('Elements that are explicitly excluded from the contract'))
    common_tos = models.TextField(blank=True)

    tracker = FieldTracker()
    objects = managers.ExpertRequestManager()

    class Meta:
        ordering = ['-created']

    def get_log_fields(self):
        return super(ExpertRequest, self).get_log_fields() + ('state', 'project', 'user')

    @classmethod
    def get_url_name(cls):
        return 'expert-request'

    @property
    def type_label(self):
        offerings = settings.WALDUR_SUPPORT.get('CONTRACT', {}).get('offerings', {})
        type_settings = offerings.get(self.type, {})
        return type_settings.get('label', None)

    @property
    def planned_budget(self):
        price = self.extra.get('price')
        if price:
            try:
                return float(price)
            except ValueError:
                return 0
        return 0

    def __str__(self):
        if self.project:
            return '%s (%s)' % (self.name, self.project)
        return self.name


class ExpertBid(core_models.UuidMixin,
                core_models.DescribableMixin,
                PriceMixin,
                structure_models.StructureLoggableMixin,
                structure_models.TimeStampedModel):
    user = models.ForeignKey(core_models.User, related_name='+', on_delete=models.CASCADE,
                             help_text=_('The user which has created this bid.'))
    request = models.ForeignKey(ExpertRequest, on_delete=models.CASCADE, related_name='bids')
    team = models.ForeignKey(structure_models.Project)
    objects = managers.ExpertBidManager()
    tracker = FieldTracker()

    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'request__project__customer'

    def get_log_fields(self):
        return super(ExpertBid, self).get_log_fields() + ('request', 'user', 'price')

    @classmethod
    def get_url_name(cls):
        return 'expert-bid'


class ExpertContract(PriceMixin, core_models.DescribableMixin, structure_models.TimeStampedModel):
    request = models.OneToOneField(ExpertRequest, on_delete=models.CASCADE, related_name='contract')
    team = models.ForeignKey(structure_models.Project, related_name='+', on_delete=models.SET_NULL, null=True)

    # Team name, team UUID and customer should be stored separately
    # because they are not available after project removal
    team_name = models.CharField(max_length=150, blank=True)
    team_uuid = models.CharField(max_length=32, blank=True)
    team_customer = models.ForeignKey(structure_models.Customer, on_delete=models.CASCADE, null=True)

    _file = models.TextField(blank=True, editable=False)

    def get_file(self):
        if not self._file:
            return

        content = base64.b64decode(self._file)
        s = StringIO.StringIO(content)
        return s

    def create_file(self):
        parser = HTMLParser()
        context = {
            'contract': self,
            'client': {
                'name': self.team.customer.name,
                'laws': self.team.customer.country,
                'number': self.team.customer.registration_code,
                'representative': self.team.customer.get_owners()[0].full_name
            },
            'expert': {
                'name': self.request.project.customer.name,
                'laws': self.request.project.customer.country,
                'number': self.request.project.customer.registration_code,
                'representative': self.request.project.customer.get_owners()[0].full_name
            },
            'service': {
                'description': parser.unescape(self.request.description),
                'date_end': ''
            },
            'status': ExpertRequest.States.COMPLETED,
            'remuneration': {
                'type': 'regular payment' if self.request.recurring_billing else 'one-time payment',
                'value': common_utils.quantize_price(self.price),
                'currency': settings.WALDUR_EXPERTS['CURRENCY_NAME']
            },
            'period': ''
        }
        contract_text = render_to_string('experts/contract_template.html', context)
        pdf = pdfkit.from_string(contract_text, False)
        self._file = base64.b64encode(pdf)
        self.save()

    def has_file(self):
        return bool(self._file)

    def get_filename(self):
        filename = "{year}_{month:02d}_{day:02d}_expert_contract_{id}.pdf".format(
            id=self.id,
            year=self.created.year,
            month=self.created.month,
            day=self.created.day
        )
        return filename

    class Meta:
        ordering = ['-created']



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-04-12 09:05
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0030_add_volume_image_name'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='internalip',
            unique_together=set([('instance', 'subnet')]),
        ),
    ]



from __future__ import unicode_literals

from django.core.validators import MaxValueValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import models as core_models
from waldur_core.core.fields import JSONField
from waldur_core.quotas.fields import CounterQuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models


class AzureService(structure_models.Service):
    Locations = (('Central US', 'Central US'),
                 ('East US 2', 'East US 2'),
                 ('South Central US', 'South Central US'),
                 ('North Europe', 'North Europe'),
                 ('East Asia', 'East Asia'),
                 ('Southeast Asia', 'Southeast Asia'),
                 ('Japan West', 'Japan West'))

    projects = models.ManyToManyField(
        structure_models.Project, related_name='azure_services', through='AzureServiceProjectLink')

    @classmethod
    def get_url_name(cls):
        return 'azure'

    class Quotas(QuotaModelMixin.Quotas):
        vm_count = CounterQuotaField(
            target_models=lambda: [VirtualMachine],
            path_to_scope='service_project_link.service'
        )


class AzureServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(AzureService)

    cloud_service_name = models.CharField(max_length=255, blank=True)

    def get_backend(self):
        return super(AzureServiceProjectLink, self).get_backend(
            cloud_service_name=self.cloud_service_name)

    @classmethod
    def get_url_name(cls):
        return 'azure-spl'


class Image(structure_models.GeneralServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'azure-image'


class Size(object):
    _meta = 'size'

    @classmethod
    def get_url_name(cls):
        return 'azure-size'


class InstanceEndpoint(core_models.BackendModelMixin, models.Model):

    class Protocol(object):
        TCP = 'tcp'
        UDP = 'udp'

        CHOICES = (
            (TCP, 'tcp'),
            (UDP, 'udp'),
        )

    class Name(object):
        SSH = 'SSH'
        RDP = 'Remote Desktop'

        CHOICES = (
            (SSH, 'SSH'),
            (RDP, 'Remote Desktop'),
        )

    local_port = models.IntegerField(validators=[MaxValueValidator(65535)])
    public_port = models.IntegerField(validators=[MaxValueValidator(65535)])
    protocol = models.CharField(max_length=3, blank=True, choices=Protocol.CHOICES)
    name = models.CharField(max_length=255, blank=True, choices=Name.CHOICES)
    instance = models.ForeignKey('VirtualMachine', related_name='endpoints', on_delete=models.PROTECT)

    @classmethod
    def get_backend_fields(cls):
        return super(InstanceEndpoint, cls).get_backend_fields() + (
            'local_port', 'public_port', 'protocol', 'name', 'vm',
        )


class VirtualMachine(structure_models.VirtualMachine):
    service_project_link = models.ForeignKey(
        AzureServiceProjectLink, related_name='virtualmachines', on_delete=models.PROTECT)
    public_ips = JSONField(default=list, help_text=_('List of public IP addresses'), blank=True)
    private_ips = JSONField(default=list, help_text=_('List of private IP addresses'), blank=True)
    user_username = models.CharField(max_length=50)
    user_password = models.CharField(max_length=50)

    @classmethod
    def get_url_name(cls):
        return 'azure-virtualmachine'

    def get_access_url_name(self):
        return 'azure-virtualmachine-rdp'

    @property
    def external_ips(self):
        return self.public_ips

    @property
    def internal_ips(self):
        return self.private_ips

    @classmethod
    def get_backend_fields(cls):
        return super(VirtualMachine, cls).get_backend_fields() + ('public_ips', 'private_ips', 'endpoints')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re

import django.contrib.auth.models
import django.core.validators
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import waldur_core.core.fields
import waldur_core.core.models


class Migration(migrations.Migration):
    replaces = [('core', '0001_initial'), ('core', '0002_user_organization_approved'),
                ('core', '0003_ssh_key_name_length_changed'),
                ('core', '0001_squashed_0003_ssh_key_name_length_changed'),
                ('core', '0002_enlarge_civil_number_user_field'), ('core', '0003_user_registration_method'),
                ('core', '0004_user_agreement_date'), ('core', '0005_add_user_language_and_competence'),
                ('core', '0006_user_is_support'), ('core', '0007_user_token_lifetime'),
                ('core', '0008_sshpublickey_is_shared')]

    initial = True

    dependencies = [
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False,
                                                     help_text='Designates that this user has all permissions without explicitly assigning them.',
                                                     verbose_name='superuser status')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('username', models.CharField(
                    help_text='Required. 30 characters or fewer. Letters, numbers and @/./+/-/_ characters',
                    max_length=30, unique=True, validators=[
                        django.core.validators.RegexValidator(re.compile('^[\\w.@+-]+$'), 'Enter a valid username.',
                                                              'invalid')], verbose_name='username')),
                ('civil_number', models.CharField(blank=True, default=None, max_length=50, null=True, unique=True,
                                                  verbose_name='civil number')),
                ('full_name', models.CharField(blank=True, max_length=100, verbose_name='full name')),
                ('native_name', models.CharField(blank=True, max_length=100, verbose_name='native name')),
                ('phone_number', models.CharField(blank=True, max_length=255, verbose_name='phone number')),
                ('organization', models.CharField(blank=True, max_length=80, verbose_name='organization')),
                ('organization_approved',
                 models.BooleanField(default=False, help_text='Designates whether user organization was approved.',
                                     verbose_name='organization approved')),
                ('job_title', models.CharField(blank=True, max_length=40, verbose_name='job title')),
                ('email', models.EmailField(blank=True, max_length=75, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False,
                                                 help_text='Designates whether the user can log into this admin site.',
                                                 verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True,
                                                  help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.',
                                                  verbose_name='active')),
                ('is_support',
                 models.BooleanField(default=False, help_text='Designates whether the user is a global support user.',
                                     verbose_name='support status')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('registration_method', models.CharField(blank=True, default='default',
                                                         help_text='Indicates what registration method were used.',
                                                         max_length=50, verbose_name='registration method')),
                ('agreement_date',
                 models.DateTimeField(blank=True, help_text='Indicates when the user has agreed with the policy.',
                                      null=True, verbose_name='agreement date')),
                ('preferred_language', models.CharField(blank=True, max_length=10)),
                ('competence', models.CharField(blank=True, max_length=255)),
                ('token_lifetime', models.PositiveIntegerField(help_text='Token lifetime in seconds.', null=True,
                                                               validators=[
                                                                   django.core.validators.MinValueValidator(60)])),
                ('groups', models.ManyToManyField(blank=True,
                                                  help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.',
                                                  related_name='user_set', related_query_name='user', to='auth.Group',
                                                  verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.',
                                                            related_name='user_set', related_query_name='user',
                                                            to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='SshPublicKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('name', models.CharField(blank=True, max_length=150)),
                ('fingerprint', models.CharField(max_length=47)),
                ('public_key', models.TextField(validators=[django.core.validators.MaxLengthValidator(2000),
                                                            waldur_core.core.models.validate_ssh_public_key])),
                ('is_shared', models.BooleanField(default=False)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'SSH public key',
                'verbose_name_plural': 'SSH public keys',
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name='sshpublickey',
            unique_together=set([('user', 'name')]),
        ),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger


class OpenStackPackageLogger(EventLogger):
    tenant = 'openstack.Tenant'
    service_settings = 'structure.ServiceSettings'
    package_template_name = basestring

    class Meta:
        event_types = ('openstack_package_created', 'openstack_package_deleted')
        event_groups = {
            'customers': event_types,
            'packages': event_types,
            'debug_only': event_types
        }
        nullable_fields = ('service_settings',)


event_logger.register('openstack_package', OpenStackPackageLogger)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-28 09:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0020_remove_external_ips'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='subnets',
            field=models.ManyToManyField(through='openstack_tenant.InternalIP', to='openstack_tenant.SubNet'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.logging.loggers
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.models
import django.db.models.deletion
import taggit.managers
import django_fsm
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('openstack_tenant', '0016_network_subnet_internalip'),
    ]

    operations = [
        migrations.CreateModel(
            name='SnapshotSchedule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('schedule', waldur_core.core.fields.CronScheduleField(max_length=15, validators=[waldur_core.core.validators.validate_cron_schedule, waldur_core.core.validators.MinCronValueValidator(1)])),
                ('next_trigger_at', models.DateTimeField(null=True)),
                ('timezone', models.CharField(default=django.utils.timezone.get_current_timezone_name, max_length=50)),
                ('is_active', models.BooleanField(default=False)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')])),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('retention_time', models.PositiveIntegerField(help_text='Retention time in days, if 0 - resource will be kept forever')),
                ('maximal_number_of_resources', models.PositiveSmallIntegerField()),
                ('call_count', models.PositiveSmallIntegerField(default=0, help_text='How many times a resource schedule was called.')),
                ('service_project_link', models.ForeignKey(related_name='snapshot_schedules', on_delete=django.db.models.deletion.PROTECT, to='openstack_tenant.OpenStackTenantServiceProjectLink')),
                ('source_volume', models.ForeignKey(related_name='snapshot_schedules', to='openstack_tenant.Volume')),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.RenameField(
            model_name='backupschedule',
            old_name='maximal_number_of_backups',
            new_name='maximal_number_of_resources',
        ),
        migrations.AddField(
            model_name='snapshot',
            name='kept_until',
            field=models.DateTimeField(help_text='Guaranteed time of snapshot retention. If null - keep forever.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='backupschedule',
            name='call_count',
            field=models.PositiveSmallIntegerField(default=0, help_text='How many times a resource schedule was called.'),
        ),
        migrations.AlterField(
            model_name='backupschedule',
            name='retention_time',
            field=models.PositiveIntegerField(help_text='Retention time in days, if 0 - resource will be kept forever'),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='snapshot_schedule',
            field=models.ForeignKey(related_name='snapshots', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='openstack_tenant.SnapshotSchedule', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-10-25 09:56
from __future__ import unicode_literals

from django.db import migrations


def populate_expert_request_initial_cache(apps, schema_editor):
    ExpertRequest = apps.get_model('experts', 'ExpertRequest')

    for request in ExpertRequest.objects.all():
        request.project_name = request.project.name
        request.project_uuid = request.project.uuid
        request.customer = request.project.customer
        request.save(update_fields=('project_name', 'project_uuid', 'customer'))


def populate_expert_contract_initial_cache(apps, schema_editor):
    ExpertContract = apps.get_model('experts', 'ExpertContract')

    for contract in ExpertContract.objects.all():
        contract.team_name = contract.team.name
        contract.team_uuid = contract.team.uuid
        contract.team_customer = contract.team.customer
        contract.save(update_fields=('team_name', 'team_uuid', 'team_customer'))


class Migration(migrations.Migration):

    dependencies = [
        ('experts', '0010_preserve_deleted_project_data'),
    ]

    operations = [
        migrations.RunPython(populate_expert_request_initial_cache),
        migrations.RunPython(populate_expert_contract_initial_cache),
    ]



from django.contrib.auth import get_user_model

from waldur_core.logging.loggers import EventLogger, event_logger


User = get_user_model()


class OpenIDEventLogger(EventLogger):
    user = User

    class Meta:
        event_types = ('auth_logged_in_with_openid',)
        event_groups = {'users': event_types}


event_logger.register('auth_openid', OpenIDEventLogger)



import logging

from django.conf import settings
from influxdb import InfluxDBClient, exceptions


logger = logging.getLogger(__name__)


def get_influxdb_client():
    if settings.WALDUR_ANALYTICS['ENABLED']:
        options = settings.WALDUR_ANALYTICS['INFLUXDB']
        return InfluxDBClient(**options)


def write_points(client, points):
    try:
        client.write_points(points)
    except (exceptions.InfluxDBClientError, exceptions.InfluxDBServerError) as e:
        logger.warning('Unable to write to InfluxDB %s', e)



from django.db import models as django_models

from waldur_core.core import managers as core_managers


class OfferingManager(core_managers.GenericKeyMixin, django_models.Manager):
    pass


class OrderItemManager(core_managers.GenericKeyMixin, django_models.Manager):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-03-07 19:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0043_servicesettings_geolocations'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicesettings',
            name='terms_of_services',
            field=models.URLField(blank=True, max_length=255),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-10-18 08:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experts', '0008_expertrequest_extra'),
    ]

    operations = [
        migrations.AlterField(
            model_name='expertrequest',
            name='description',
            field=models.TextField(blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-07-14 06:31
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('experts', '0003_expertbid'),
    ]

    operations = [
        migrations.AddField(
            model_name='expertbid',
            name='description',
            field=models.CharField(blank=True, max_length=500, verbose_name='description'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-08 10:06
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0013_add_more_backend_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='backend_id',
            field=models.CharField(max_length=255, unique=True),
        ),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger
from .models import Invoice, Payment


class InvoiceEventLogger(EventLogger):
    invoice = Invoice

    class Meta:
        event_types = ('invoice_deletion_succeeded',
                       'invoice_update_succeeded',
                       'invoice_creation_succeeded')
        event_groups = {'invoices': event_types}


class PaymentEventLogger(EventLogger):
    payment = Payment

    class Meta:
        event_types = ('payment_creation_succeeded',
                       'payment_approval_succeeded',
                       'payment_cancel_succeeded')
        event_groups = {'payments': event_types}


event_logger.register('paypal_invoice', InvoiceEventLogger)
event_logger.register('paypal_payment', PaymentEventLogger)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('logging', '0006_pushhook_systemnotification'),
    ]

    operations = [
        migrations.RenameField(
            model_name='pushhook',
            old_name='registration_token',
            new_name='device_id',
        ),
        migrations.AlterField(
            model_name='pushhook',
            name='device_id',
            field=models.CharField(max_length=255, unique=True, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='pushhook',
            unique_together=set([('user', 'device_id', 'type')]),
        ),
    ]



from __future__ import unicode_literals

from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status, exceptions as rf_exceptions
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from waldur_core.core import views as core_views, validators as core_validators
from waldur_core.core.mixins import EagerLoadMixin
from waldur_core.structure import permissions as structure_permissions, filters as structure_filters

from . import serializers, models, filters


class BaseMarketplaceView(core_views.ActionsViewSet):
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    update_permissions = \
        partial_update_permissions = \
        destroy_permissions = \
        [structure_permissions.is_owner]


class ServiceProviderViewSet(BaseMarketplaceView):
    queryset = models.ServiceProvider.objects.all()
    serializer_class = serializers.ServiceProviderSerializer
    filter_class = filters.ServiceProviderFilter


class CategoryViewSet(EagerLoadMixin, core_views.ActionsViewSet):
    queryset = models.Category.objects.all()
    serializer_class = serializers.CategorySerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)

    create_permissions = \
        update_permissions = \
        partial_update_permissions = \
        destroy_permissions = \
        [structure_permissions.is_staff]


class OfferingViewSet(BaseMarketplaceView):
    queryset = models.Offering.objects.all()
    serializer_class = serializers.OfferingSerializer
    filter_class = filters.OfferingFilter


class PlanViewSet(BaseMarketplaceView):
    queryset = models.Plan.objects.all()
    serializer_class = serializers.PlanSerializer
    filter_class = filters.PlanFilter


class ScreenshotViewSet(BaseMarketplaceView):
    queryset = models.Screenshots.objects.all()
    serializer_class = serializers.ScreenshotSerializer
    filter_class = filters.ScreenshotFilter


class OrderViewSet(BaseMarketplaceView):
    queryset = models.Order.objects.all()
    serializer_class = serializers.OrderSerializer
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.OrderFilter

    def check_permissions_for_state_change(request, view, order=None):
        if not order:
            return

        user = request.user

        if user.is_staff:
            return

        if settings.WALDUR_MARKETPLACE['OWNER_CAN_APPROVE_ORDER'] and \
                structure_permissions._has_owner_access(user, order.project.customer):
            return

        if settings.WALDUR_MARKETPLACE['MANAGER_CAN_APPROVE_ORDER'] and \
                structure_permissions._has_manager_access(user, order.project):
            return

        if settings.WALDUR_MARKETPLACE['ADMIN_CAN_APPROVE_ORDER'] and \
                structure_permissions._has_admin_access(user, order.project):
            return

        raise rf_exceptions.PermissionDenied()

    @detail_route(methods=['post'])
    def set_state_requested_for_approval(self, request, uuid=None):
        return self._update_state(request, models.Order.States.REQUESTED_FOR_APPROVAL)

    set_state_requested_for_approval_validators = [core_validators.StateValidator(models.Order.States.DRAFT)]

    @detail_route(methods=['post'])
    def set_state_executing(self, request, uuid=None):
        order = self.get_object()
        for item in order.items.all():
            item.process(request)
        return self._update_state(request, models.Order.States.EXECUTING)

    set_state_executing_validators = [core_validators.StateValidator(models.Order.States.REQUESTED_FOR_APPROVAL)]
    set_state_executing_permissions = [check_permissions_for_state_change]

    @detail_route(methods=['post'])
    def set_state_done(self, request, uuid=None):
        order = self.get_object()
        response = self._update_state(request, models.Order.States.DONE, order)
        order.approved_by = request.user
        order.approved_at = timezone.now()
        order.save()
        return response

    set_state_done_validators = [core_validators.StateValidator(models.Order.States.EXECUTING)]
    set_state_done_permissions = [check_permissions_for_state_change]

    @detail_route(methods=['post'])
    def set_state_terminated(self, request, uuid=None):
        return self._update_state(request, models.Order.States.TERMINATED)

    def _update_state(self, request, state, order=None):
        if not order:
            order = self.get_object()

        state_name = filter(lambda x: x[0] == state, models.Order.States.CHOICES)[0][1]
        state_name = state_name.replace(' ', '_')
        getattr(order, 'set_state_' + state_name)()
        order.save(update_fields=['state'])
        return Response({'detail': _('Order state updated.')},
                        status=status.HTTP_200_OK)



from waldur_core.cost_tracking import CostTrackingStrategy, ConsumableItem, CostTrackingRegister

from . import models, utils, PriceItemTypes


class InstanceStrategy(CostTrackingStrategy):
    resource_class = models.Instance

    class Types(object):
        FLAVOR = PriceItemTypes.FLAVOR

    @classmethod
    def get_consumable_items(cls):
        for flavor_name in set(models.Flavor.objects.all().values_list('name', flat=True)):
            yield utils.get_consumable_item(flavor_name)

    @classmethod
    def get_configuration(cls, instance):
        consumables = {}
        if instance.state != models.Instance.States.ERRED:
            consumables[ConsumableItem(item_type=cls.Types.FLAVOR, key=instance.flavor_name)] = 1
        return consumables


CostTrackingRegister.register_strategy(InstanceStrategy)


class VolumeStrategy(CostTrackingStrategy):
    resource_class = models.Volume

    class Types(object):
        STORAGE = PriceItemTypes.STORAGE

    class Keys(object):
        STORAGE = '1 GB'

    @classmethod
    def get_consumable_items(cls):
        return [ConsumableItem(item_type=cls.Types.STORAGE, key=cls.Keys.STORAGE, name='1 GB of storage', units='GB')]

    @classmethod
    def get_configuration(cls, volume):
        return {ConsumableItem(item_type=cls.Types.STORAGE, key=cls.Keys.STORAGE): float(volume.size) / 1024}


CostTrackingRegister.register_strategy(VolumeStrategy)


class SnapshotStrategy(CostTrackingStrategy):
    resource_class = models.Snapshot

    class Types(object):
        STORAGE = PriceItemTypes.STORAGE

    class Keys(object):
        STORAGE = '1 GB'

    @classmethod
    def get_consumable_items(cls):
        return [ConsumableItem(item_type=cls.Types.STORAGE, key=cls.Keys.STORAGE, name='1 GB of storage', units='GB')]

    @classmethod
    def get_configuration(cls, snapshot):
        return {ConsumableItem(item_type=cls.Types.STORAGE, key=cls.Keys.STORAGE): float(snapshot.size) / 1024}


CostTrackingRegister.register_strategy(SnapshotStrategy)



from __future__ import unicode_literals

import six

from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _


class AttributeType(object):
    @staticmethod
    def available_values_validate(values):
        if values:
            raise ValidationError(_('Available values must be empty for this attribute type.'))

    @staticmethod
    def validate(values, available_values):
        raise NotImplementedError


class BooleanAttribute(AttributeType):
    @staticmethod
    def validate(values, available_values=None):
        if not isinstance(values, bool):
            raise ValidationError(_('Value must be a boolean type  for this attribute type.'))


class StringAttribute(AttributeType):
    @staticmethod
    def validate(values, available_values=None):
        if not isinstance(values, six.text_type):
            raise ValidationError(_('Value must be a boolean type  for this attribute type.'))


class IntegerAttribute(AttributeType):
    @staticmethod
    def validate(values, available_values=None):
        if isinstance(values, bool) or not isinstance(values, int):
            raise ValidationError(_('Value must be an integer type  for this attribute type.'))


class ChoiceAttribute(AttributeType):
    @staticmethod
    def available_values_validate(values):
        if not values:
            raise ValidationError(_('Available values not must be empty for this attribute type.'))

        if not isinstance(values, list):
            raise ValidationError(_('Available values must be a list for this attribute type.'))

    @staticmethod
    def validate(values, available_values):
        if not isinstance(values, six.text_type):
            raise ValidationError(_('Value must be a string for this attribute.'))

        if not(values in available_values):
            raise ValidationError(_('This value is not available for this attribute.'))


class ListAttribute(AttributeType):
    @staticmethod
    def available_values_validate(values):
        if not values:
            raise ValidationError(_('Available values not must be empty for this attribute type.'))

        if not isinstance(values, list):
            raise ValidationError(_('Available values must be a list for this attribute type.'))

    @staticmethod
    def validate(values, available_values):
        if not isinstance(values, list):
            raise ValidationError(_('Value must be a list for this attribute.'))

        if not(set(values) <= set(available_values)):
            raise ValidationError(_('These values are not available for this attribute.'))


ATTRIBUTE_TYPES = (
    ('boolean', 'boolean'),
    ('string', 'string'),
    ('integer', 'integer'),
    ('choice', 'choice'),
    ('list', 'list'),
)


def get_attribute_type(name):
    attribute_type = {
        'boolean': BooleanAttribute,
        'string': StringAttribute,
        'integer': IntegerAttribute,
        'choice': ChoiceAttribute,
        'list': ListAttribute,
    }
    try:
        return attribute_type[name]
    except KeyError:
        pass



from __future__ import unicode_literals

from datetime import timedelta
import logging

from django.conf import settings
from django.db.models import Q
from django.db import transaction
from django.utils import timezone

from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.core import models as core_models
from waldur_core.quotas import exceptions as quotas_exceptions
from waldur_core.structure import (models as structure_models, tasks as structure_tasks,
                                   SupportedServices)

from . import models, serializers, log


logger = logging.getLogger(__name__)


class SetInstanceOKTask(core_tasks.StateTransitionTask):
    """ Additionally mark or related floating IPs as free """

    def pre_execute(self, instance):
        self.kwargs['state_transition'] = 'set_ok'
        self.kwargs['action'] = ''
        self.kwargs['action_details'] = {}
        super(SetInstanceOKTask, self).pre_execute(instance)

    def execute(self, instance, *args, **kwargs):
        super(SetInstanceOKTask, self).execute(instance)
        instance.floating_ips.update(is_booked=False)


class SetInstanceErredTask(core_tasks.ErrorStateTransitionTask):
    """ Mark instance as erred and delete resources that were not created. """

    def execute(self, instance):
        super(SetInstanceErredTask, self).execute(instance)

        # delete volumes if they were not created on backend,
        # mark as erred if creation was started, but not ended,
        # leave as is, if they are OK.
        for volume in instance.volumes.all():
            if volume.state == models.Volume.States.CREATION_SCHEDULED:
                volume.delete()
            elif volume.state == models.Volume.States.OK:
                pass
            else:
                volume.set_erred()
                volume.save(update_fields=['state'])

        # set instance floating IPs as free, delete not created ones.
        instance.floating_ips.filter(backend_id='').delete()
        instance.floating_ips.update(is_booked=False)


class SetBackupErredTask(core_tasks.ErrorStateTransitionTask):
    """ Mark DR backup and all related resources that are not in state OK as Erred """

    def execute(self, backup):
        super(SetBackupErredTask, self).execute(backup)
        for snapshot in backup.snapshots.all():
            # If snapshot creation was not started - delete it from NC DB.
            if snapshot.state == models.Snapshot.States.CREATION_SCHEDULED:
                snapshot.decrease_backend_quotas_usage()
                snapshot.delete()
            else:
                snapshot.set_erred()
                snapshot.save(update_fields=['state'])

        # Deactivate schedule if its backup become erred.
        schedule = backup.backup_schedule
        if schedule:
            schedule.error_message = 'Failed to execute backup schedule for %s. Error: %s' % (
                backup.instance, backup.error_message)
            schedule.is_active = False
            schedule.save()


class ForceDeleteBackupTask(core_tasks.DeletionTask):

    def execute(self, backup):
        backup.snapshots.all().delete()
        super(ForceDeleteBackupTask, self).execute(backup)


class VolumeExtendErredTask(core_tasks.ErrorStateTransitionTask):
    """ Mark volume and its instance as erred on fail """

    def execute(self, volume):
        super(VolumeExtendErredTask, self).execute(volume)
        if volume.instance is not None:
            super(VolumeExtendErredTask, self).execute(volume.instance)


class BaseScheduleTask(core_tasks.BackgroundTask):
    """
    This task has several important caveats to consider.

    1. If user has decreased value of maximal_number_of_resources attribute,
       but exceeding resources have been already created, we would try to automatically delete
       exceeding resources before creating new resources.
       However, if new resource creation fails, old resources cannot be restored.

       Therefore, it is strongly advised to modify value of maximal_number_of_resources attribute very carefully.
       Also it is better to delete exceeding resources manually instead of relying on automatic deletion
       so that it is easier to explicitly select resources to be removed.

    2. _remove_exceeding_resources method orders resources by value of kept_until attribute in ASC order.
       It assumes that NULL values come *last* with ascending sort order.
       Therefore it would work correctly only in PostgreSQL.
       It would not work correctly in MySQL because in MySQL NULL values come *first*.

    3. Value of kept_until attribute is ignored as long as there are exceeding resources.
       It means that existing resources are deleted even if it is requested to be kept forever.
       Essentially, retention_time and maximal_number_of_resources attributes are mutually exclusive.

       Consider, for example, case when value of maximal_number_of_resources is 3 and there are 6 resources,
       out of which 2 with non-null value of kept_until attribute and 4 resources to be kept forever.
       As you can see, there are 3 exceeding resources, which should be removed.

       Then, both 2 first resources would be deleted, and 1 resource to be kept forever is deleted as well.
       Please note that last resource for deletion is chosen by value of *created* attribute.
       It means that oldest resource is selected for deletion.

    4. Database records for resources are created and deleted synchronously,
       but actual backend API task are scheduled asynchronously.
       Therefore, next iteration of schedule task does not wait
       until previous iteration tasks are completed.
       That's why there may several concurrent execution of the same schedule.

    5. Actual execution of schedule depends on number of Celery workers and their load.
       For example, even if schedule is expected to create new resources each hour,
       but all Celery workers have been overloaded for 2 hours, only one resource would be created.

    6. Schedule is disabled as long as resource quota is exceeded.
       Schedule is not reactivated automatically whenever quota limit
       is increased or quota usage is decreased.
       Instead it is expected that user would manually reactivate schedule in this case.

    7. Schedule is skipped and new resources are not created as long as schedule is disabled or
       number of resources has reached value of maximal_number_of_resources attribute.
    """

    model = NotImplemented
    resource_attribute = NotImplemented

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    @transaction.atomic()
    def run(self):
        schedules = self.model.objects.filter(is_active=True, next_trigger_at__lt=timezone.now())
        for schedule in schedules:
            existing_resources = self._get_number_of_resources(schedule)
            if existing_resources > schedule.maximal_number_of_resources:
                self._schedule_exceeding_resources_deletion(schedule, existing_resources)
                continue
            elif existing_resources == schedule.maximal_number_of_resources:
                logger.debug('Skipping schedule %s because number of resources %s has reached limit %s.',
                             schedule, existing_resources, schedule.maximal_number_of_resources)
                continue

            kept_until = None
            if schedule.retention_time:
                kept_until = timezone.now() + timezone.timedelta(days=schedule.retention_time)

            try:
                # Value of call_count attribute is used as suffix of new resource name
                schedule.call_count += 1
                schedule.save()
                resource = self._create_resource(schedule, kept_until=kept_until)
            except quotas_exceptions.QuotaValidationError as e:
                message = 'Failed to schedule "%s" creation. Error: %s' % (self.model.__name__, e)
                logger.exception(
                    'Resource schedule (PK: %s), (Name: %s) execution failed. %s' % (schedule.pk,
                                                                                     schedule.name,
                                                                                     message))
                schedule.is_active = False
                schedule.error_message = message
                schedule.save()
            else:
                executor = self._get_create_executor()
                executor.execute(resource)
                schedule.update_next_trigger_at()
                schedule.save()

    def _schedule_exceeding_resources_deletion(self, schedule, resources_count):
        amount_to_remove = resources_count - schedule.maximal_number_of_resources
        self._log_backup_cleanup(schedule, amount_to_remove, resources_count)
        ok_or_erred = Q(state=core_models.StateMixin.States.OK) | Q(state=core_models.StateMixin.States.ERRED)
        queryset = getattr(schedule, self.resource_attribute)
        resources = queryset.filter(ok_or_erred).order_by('kept_until', 'created')
        resources_to_remove = resources[:amount_to_remove]
        executor = self._get_delete_executor()
        for resource in resources_to_remove:
            executor.execute(resource)

    def _log_backup_cleanup(self, schedule, amount_to_remove, resources_count):
        raise NotImplementedError()

    def _create_resource(self, schedule, kept_until):
        raise NotImplementedError()

    def _get_create_executor(self):
        raise NotImplementedError()

    def _get_delete_executor(self):
        raise NotImplementedError()

    def _get_number_of_resources(self, schedule):
        resources = getattr(schedule, self.resource_attribute)
        return resources.count()


class ScheduleBackups(BaseScheduleTask):
    name = 'openstack_tenant.ScheduleBackups'
    model = models.BackupSchedule
    resource_attribute = 'backups'

    def _create_resource(self, schedule, kept_until):
        backup = models.Backup.objects.create(
            name='Backup#%s of %s' % (schedule.call_count, schedule.instance.name),
            description='Scheduled backup of instance "%s"' % schedule.instance,
            service_project_link=schedule.instance.service_project_link,
            instance=schedule.instance,
            backup_schedule=schedule,
            metadata=serializers.BackupSerializer.get_backup_metadata(schedule.instance),
            kept_until=kept_until,
        )
        serializers.BackupSerializer.create_backup_snapshots(backup)
        return backup

    def _get_create_executor(self):
        from . import executors
        return executors.BackupCreateExecutor

    def _get_delete_executor(self):
        from . import executors
        return executors.BackupDeleteExecutor

    def _log_backup_cleanup(self, schedule, amount_to_remove, resources_count):
        message_template = ('Maximum resource count "%s" has been reached.'
                            '"%s" from "%s" resources are going to be removed.')
        log.event_logger.openstack_backup_schedule.info(
            message_template % (schedule.maximal_number_of_resources, amount_to_remove, resources_count),
            event_type='resource_backup_schedule_cleaned_up',
            event_context={'resource': schedule.instance, 'backup_schedule': schedule},
        )


class BaseDeleteExpiredResourcesTask(core_tasks.BackgroundTask):
    model = NotImplemented

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    def _get_executor(self):
        raise NotImplementedError()

    @transaction.atomic
    def run(self):
        executor = self._get_delete_executor()
        resources = self.model.objects.filter(kept_until__lt=timezone.now(),
                                              state=core_models.StateMixin.States.OK)
        for resource in resources:
            executor.execute(resource)


class DeleteExpiredBackups(BaseDeleteExpiredResourcesTask):
    name = 'openstack_tenant.DeleteExpiredBackups'
    model = models.Backup

    def _get_delete_executor(self):
        from . import executors
        return executors.BackupDeleteExecutor


class ScheduleSnapshots(BaseScheduleTask):
    name = 'openstack_tenant.ScheduleSnapshots'
    model = models.SnapshotSchedule
    resource_attribute = 'snapshots'

    def _create_resource(self, schedule, kept_until):
        snapshot = models.Snapshot.objects.create(
            name='Snapshot#%s of %s' % (schedule.call_count, schedule.source_volume.name),
            description='Scheduled snapshot of volume "%s"' % schedule.source_volume,
            service_project_link=schedule.source_volume.service_project_link,
            source_volume=schedule.source_volume,
            snapshot_schedule=schedule,
            size=schedule.source_volume.size,
            metadata=serializers.SnapshotSerializer.get_snapshot_metadata(schedule.source_volume),
            kept_until=kept_until,
        )
        snapshot.increase_backend_quotas_usage()
        return snapshot

    def _get_create_executor(self):
        from . import executors
        return executors.SnapshotCreateExecutor

    def _get_delete_executor(self):
        from . import executors
        return executors.SnapshotDeleteExecutor

    def _log_backup_cleanup(self, schedule, amount_to_remove, resources_count):
        message_template = ('Maximum resource count "%s" has been reached.'
                            '"%s" from "%s" resources are going to be removed.')
        log.event_logger.openstack_snapshot_schedule.info(
            message_template % (schedule.maximal_number_of_resources, amount_to_remove, resources_count),
            event_type='resource_snapshot_schedule_cleaned_up',
            event_context={'resource': schedule.source_volume, 'snapshot_schedule': schedule},
        )


class DeleteExpiredSnapshots(BaseDeleteExpiredResourcesTask):
    name = 'openstack_tenant.DeleteExpiredSnapshots'
    model = models.Snapshot

    def _get_delete_executor(self):
        from . import executors
        return executors.SnapshotDeleteExecutor


class SetErredStuckResources(core_tasks.BackgroundTask):
    name = 'openstack_tenant.SetErredStuckResources'

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    def run(self):
        for model in (models.Instance, models.Volume, models.Snapshot):
            cutoff = timezone.now() - timedelta(minutes=30)
            for resource in model.objects.filter(modified__lt=cutoff,
                                                 state=structure_models.NewResource.States.CREATING):
                resource.set_erred()
                resource.error_message = 'Provisioning is timed out.'
                resource.save(update_fields=['state', 'error_message'])
                logger.warning('Switching resource %s to erred state, '
                               'because provisioning is timed out.',
                               core_utils.serialize_instance(resource))


class LimitedPerTypeThrottleMixin(object):

    def get_limit(self, resource):
        nc_settings = getattr(settings, 'WALDUR_OPENSTACK', {})
        limit_per_type = nc_settings.get('MAX_CONCURRENT_PROVISION', {})
        model_name = SupportedServices.get_name_for_model(resource)
        return limit_per_type.get(model_name, super(LimitedPerTypeThrottleMixin, self).get_limit(resource))


class ThrottleProvisionTask(LimitedPerTypeThrottleMixin, structure_tasks.ThrottleProvisionTask):
    pass


class ThrottleProvisionStateTask(LimitedPerTypeThrottleMixin, structure_tasks.ThrottleProvisionStateTask):
    pass



from waldur_core.structure import filters as structure_filters

from . import models


class ImageFilter(structure_filters.BaseServicePropertyFilter):
    class Meta(structure_filters.BaseServicePropertyFilter.Meta):
        model = models.Image



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0014_make_backupschedule_spl_non_nullable'),
    ]

    operations = [
        migrations.CreateModel(
            name='SnapshotRestoration',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('snapshot', models.ForeignKey(related_name='restorations', to='openstack_tenant.Snapshot')),
                ('volume', models.OneToOneField(related_name='restoration', to='openstack_tenant.Volume')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist

from waldur_mastermind.packages import models as package_models
from waldur_mastermind.marketplace import models as marketplace_models


logger = logging.getLogger(__name__)


def create_offering_for_package_template(template):
    conf = settings.WALDUR_MARKETPLACE_PACKAGES
    customer_id = conf.get('CUSTOMER_ID')
    category_id = conf.get('CATEGORY_ID')

    if not customer_id:
        logger.debug('Marketplace offering for VPC is not '
                     'created because customer ID is not defined.')
        return

    if not category_id:
        logger.debug('Marketplace offering for VPC is not '
                     'created because category ID is not defined.')
        return

    attributes = {
        component.type: component.amount
        for component in template.components.all()
    }
    marketplace_models.Offering.objects.create(
        name=template.name,
        description=template.description,
        customer_id=customer_id,
        category_id=category_id,
        geolocations=template.service_settings.geolocations,
        attributes=attributes,
        scope=template,
    )


def update_offering_for_template(template):
    fields = {}
    changed = set(template.tracker.changed())
    if 'name' in changed:
        fields['name'] = template.name
    if 'description' in changed:
        fields['description'] = template.description
    if 'archived' in changed:
        fields['is_active'] = not template.archived
    if fields:
        marketplace_models.Offering.objects.filter(scope=template).update(**fields)


def sync_offering_attribute_with_template_component(component):
    try:
        offering = marketplace_models.Offering.objects.get(scope=component.template)
    except ObjectDoesNotExist:
        logger.debug('Skipping offering attributes synchronization because offering is not found.')
    else:
        offering.attributes[component.type] = component.amount
        offering.save(update_fields=['attributes'])


def create_missing_offerings():
    content_type = ContentType.objects.get_for_model(package_models.PackageTemplate)
    offerings = marketplace_models.Offering.objects.filter(content_type=content_type)
    front_templates = set(offerings.values_list('object_id', flat=True))
    back_templates = set(package_models.PackageTemplate.objects.all().values_list('pk', flat=True))
    missing_ids = back_templates - front_templates

    missing_templates = package_models.PackageTemplate.objects.filter(pk__in=missing_ids)
    for template in missing_templates:
        create_offering_for_package_template(template)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-04-06 08:51
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jupyter_hub_management', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='jupyterhubmanagement',
            name='python_management',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='jupyter_hub_management', to='python_management.PythonManagement'),
        ),
    ]



from celery import chain

from waldur_core.core import executors

from .tasks import AuthTask, PollTask


class AuthExecutor(executors.ErrorExecutorMixin, executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance):
        return chain(
            AuthTask().si(serialized_instance, state_transition='begin_processing'),
            PollTask().si(serialized_instance).set(countdown=30),
        )



from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, exceptions, response, status, serializers as rf_serializers

from waldur_core.core import validators as core_validators, exceptions as core_exceptions
from waldur_core.structure import (views as structure_views, filters as structure_filters,
                                   permissions as structure_permissions)

from . import models, filters, serializers, executors


class OpenStackServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.OpenStackService.objects.all().order_by('id')
    serializer_class = serializers.ServiceSerializer

    def list(self, request, *args, **kwargs):
        """
        To create a service, issue a **POST** to */api/openstack/* as a customer owner.

        You can create service based on shared service settings. Example:

        .. code-block:: http

            POST /api/openstack/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "Common OpenStack",
                "customer": "http://example.com/api/customers/1040561ca9e046d2b74268600c7e1105/",
                "settings": "http://example.com/api/service-settings/93ba615d6111466ebe3f792669059cb4/"
            }

        Or provide your own credentials. Example:

        .. code-block:: http

            POST /api/openstack/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "My OpenStack",
                "customer": "http://example.com/api/customers/1040561ca9e046d2b74268600c7e1105/",
                "backend_url": "http://keystone.example.com:5000/v2.0",
                "username": "admin",
                "password": "secret"
            }
        """

        return super(OpenStackServiceViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        To update OpenStack service issue **PUT** or **PATCH** against */api/openstack/<service_uuid>/*
        as a customer owner. You can update service's `name` and `available_for_all` fields.

        Example of a request:

        .. code-block:: http

            PUT /api/openstack/c6526bac12b343a9a65c4cd6710666ee/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "name": "My OpenStack2"
            }

        To remove OpenStack service, issue **DELETE** against */api/openstack/<service_uuid>/* as
        staff user or customer owner.
        """
        return super(OpenStackServiceViewSet, self).retrieve(request, *args, **kwargs)


class OpenStackServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.OpenStackServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filter_class = filters.OpenStackServiceProjectLinkFilter

    def list(self, request, *args, **kwargs):
        """
        In order to be able to provision OpenStack resources, it must first be linked to a project. To do that,
        **POST** a connection between project and a service to */api/openstack-service-project-link/*
        as stuff user or customer owner.

        Example of a request:

        .. code-block:: http

            POST /api/openstack-service-project-link/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "project": "http://example.com/api/projects/e5f973af2eb14d2d8c38d62bcbaccb33/",
                "service": "http://example.com/api/openstack/b0e8a4cbd47c4f9ca01642b7ec033db4/"
            }

        To remove a link, issue DELETE to URL of the corresponding connection as stuff user or customer owner.
        """
        return super(OpenStackServiceProjectLinkViewSet, self).list(request, *args, **kwargs)


class FlavorViewSet(structure_views.BaseServicePropertyViewSet):
    """
    VM instance flavor is a pre-defined set of virtual hardware parameters that the instance will use:
    CPU, memory, disk size etc. VM instance flavor is not to be confused with VM template -- flavor is a set of virtual
    hardware parameters whereas template is a definition of a system to be installed on this instance.
    """
    queryset = models.Flavor.objects.all().order_by('settings', 'cores', 'ram', 'disk')
    serializer_class = serializers.FlavorSerializer
    lookup_field = 'uuid'
    filter_class = filters.FlavorFilter


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    lookup_field = 'uuid'
    filter_class = filters.ImageFilter


class SecurityGroupViewSet(structure_views.BaseResourceViewSet):
    queryset = models.SecurityGroup.objects.all()
    serializer_class = serializers.SecurityGroupSerializer
    filter_class = filters.SecurityGroupFilter
    disabled_actions = ['create', 'pull']  # pull operation should be implemented in WAL-323

    def default_security_group_validator(security_group):
        if security_group.name == 'default':
            raise exceptions.ValidationError({
                'name': _('Default security group is managed by OpenStack itself.')
            })

    update_validators = partial_update_validators = structure_views.ResourceViewSet.update_validators + [
        default_security_group_validator
    ]
    update_executor = executors.SecurityGroupUpdateExecutor

    destroy_validators = structure_views.ResourceViewSet.destroy_validators + [
        default_security_group_validator
    ]
    delete_executor = executors.SecurityGroupDeleteExecutor

    @decorators.detail_route(methods=['POST'])
    def set_rules(self, request, uuid=None):
        """ WARNING! Auto-generated HTML form is wrong for this endpoint. List should be defined as input.

            Example:
            [
                {
                    "protocol": "tcp",
                    "from_port": 1,
                    "to_port": 10,
                    "cidr": "10.1.1.0/24"
                }
            ]
        """
        # XXX: DRF does not support forms generation for list serializers.
        #      Thats why we use different serializer in view.
        serializer = serializers.SecurityGroupRuleListUpdateSerializer(
            data=request.data, context=self.get_serializer_context())
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.PushSecurityGroupRulesExecutor().execute(self.get_object())
        return response.Response(
            {'status': _('Rules update was successfully scheduled.')}, status=status.HTTP_202_ACCEPTED)

    set_rules_validators = [core_validators.StateValidator(models.Tenant.States.OK)]
    set_rules_serializer_class = serializers.SecurityGroupRuleUpdateSerializer


class FloatingIPViewSet(structure_views.BaseResourceViewSet):
    queryset = models.FloatingIP.objects.all().order_by('address')
    serializer_class = serializers.FloatingIPSerializer
    filter_class = filters.FloatingIPFilter
    disabled_actions = ['update', 'partial_update', 'create']
    delete_executor = executors.FloatingIPDeleteExecutor
    pull_executor = executors.FloatingIPPullExecutor

    def list(self, request, *args, **kwargs):
        """
        To get a list of all available floating IPs, issue **GET** against */api/floating-ips/*.
        Floating IPs are read only. Each floating IP has fields: 'address', 'status'.

        Status *DOWN* means that floating IP is not linked to a VM, status *ACTIVE* means that it is in use.
        """

        return super(FloatingIPViewSet, self).list(request, *args, **kwargs)


class TenantViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Tenant.objects.all()
    serializer_class = serializers.TenantSerializer
    filter_class = structure_filters.BaseResourceFilter

    create_executor = executors.TenantCreateExecutor
    update_executor = executors.TenantUpdateExecutor
    pull_executor = executors.TenantPullExecutor

    importable_resources_backend_method = 'get_tenants_for_import'
    importable_resources_serializer_class = serializers.TenantImportableSerializer
    importable_resources_permissions = [structure_permissions.is_staff]
    import_resource_serializer_class = serializers.TenantImportSerializer
    import_resource_permissions = [structure_permissions.is_staff]
    import_resource_executor = executors.TenantImportExecutor

    def delete_permission_check(request, view, obj=None):
        if not obj:
            return
        if obj.service_project_link.service.settings.shared:
            if settings.WALDUR_OPENSTACK['MANAGER_CAN_MANAGE_TENANTS']:
                structure_permissions.is_manager(request, view, obj)
            else:
                structure_permissions.is_owner(request, view, obj)
        else:
            structure_permissions.is_administrator(request, view, obj)

    delete_executor = executors.TenantDeleteExecutor
    destroy_permissions = [
        delete_permission_check,
        structure_permissions.check_access_to_services_management,
    ]
    create_permissions = update_permissions = partial_update_permissions = [
        structure_permissions.check_access_to_services_management,
    ]

    @decorators.detail_route(methods=['post'])
    def set_quotas(self, request, uuid=None):
        """
        A quota can be set for a particular tenant. Only staff users can do that.
        In order to set quota submit **POST** request to */api/openstack-tenants/<uuid>/set_quotas/*.
        The quota values are propagated to the backend.

        The following quotas are supported. All values are expected to be integers:

        - instances - maximal number of created instances.
        - ram - maximal size of ram for allocation. In MiB_.
        - storage - maximal size of storage for allocation. In MiB_.
        - vcpu - maximal number of virtual cores for allocation.
        - security_group_count - maximal number of created security groups.
        - security_group_rule_count - maximal number of created security groups rules.
        - volumes - maximal number of created volumes.
        - snapshots - maximal number of created snapshots.

        It is possible to update quotas by one or by submitting all the fields in one request.
        Waldur will attempt to update the provided quotas. Please note, that if provided quotas are
        conflicting with the backend (e.g. requested number of instances is below of the already existing ones),
        some quotas might not be applied.

        .. _MiB: http://en.wikipedia.org/wiki/Mebibyte

        Example of a valid request (token is user specific):

        .. code-block:: http

            POST /api/openstack-tenants/c84d653b9ec92c6cbac41c706593e66f567a7fa4/set_quotas/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Host: example.com

            {
                "instances": 30,
                "ram": 100000,
                "storage": 1000000,
                "vcpu": 30,
                "security_group_count": 100,
                "security_group_rule_count": 100,
                "volumes": 10,
                "snapshots": 20
            }

        Response code of a successful request is **202 ACCEPTED**.
        In case tenant is in a non-stable status, the response would be **409 CONFLICT**.
        In this case REST client is advised to repeat the request after some time.
        On successful completion the task will synchronize quotas with the backend.
        """
        tenant = self.get_object()

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        quotas = dict(serializer.validated_data)
        for quota_name, limit in quotas.items():
            tenant.set_quota_limit(quota_name, limit)
        executors.TenantPushQuotasExecutor.execute(tenant, quotas=quotas)

        return response.Response(
            {'detail': _('Quota update has been scheduled')}, status=status.HTTP_202_ACCEPTED)

    set_quotas_permissions = [structure_permissions.is_staff]
    set_quotas_validators = [core_validators.StateValidator(models.Tenant.States.OK)]
    set_quotas_serializer_class = serializers.TenantQuotaSerializer

    @decorators.detail_route(methods=['post'])
    def create_network(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        network = serializer.save()

        executors.NetworkCreateExecutor().execute(network)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_network_validators = [core_validators.StateValidator(models.Tenant.States.OK)]
    create_network_serializer_class = serializers.NetworkSerializer

    def external_network_is_defined(tenant):
        if not tenant.external_network_id:
            raise core_exceptions.IncorrectStateException(
                _('Cannot create floating IP if tenant external network is not defined.'))

    @decorators.detail_route(methods=['post'])
    def create_floating_ip(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        floating_ip = serializer.save()

        executors.FloatingIPCreateExecutor.execute(floating_ip)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_floating_ip_validators = [core_validators.StateValidator(models.Tenant.States.OK),
                                     external_network_is_defined]
    create_floating_ip_serializer_class = serializers.FloatingIPSerializer

    @decorators.detail_route(methods=['post'])
    def pull_floating_ips(self, request, uuid=None):
        tenant = self.get_object()

        executors.TenantPullFloatingIPsExecutor.execute(tenant)
        return response.Response(status=status.HTTP_202_ACCEPTED)

    pull_floating_ips_validators = [core_validators.StateValidator(models.Tenant.States.OK)]
    pull_floating_ips_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def create_security_group(self, request, uuid=None):
        """
        Example of a request:

        .. code-block:: http

            {
                "name": "Security group name",
                "description": "description",
                "rules": [
                    {
                        "protocol": "tcp",
                        "from_port": 1,
                        "to_port": 10,
                        "cidr": "10.1.1.0/24"
                    },
                    {
                        "protocol": "udp",
                        "from_port": 10,
                        "to_port": 8000,
                        "cidr": "10.1.1.0/24"
                    }
                ]
            }
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        security_group = serializer.save()

        executors.SecurityGroupCreateExecutor().execute(security_group)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_security_group_validators = [core_validators.StateValidator(models.Tenant.States.OK)]
    create_security_group_serializer_class = serializers.SecurityGroupSerializer

    @decorators.detail_route(methods=['post'])
    def pull_security_groups(self, request, uuid=None):
        executors.TenantPullSecurityGroupsExecutor.execute(self.get_object())
        return response.Response(
            {'status': _('Security groups pull has been scheduled.')}, status=status.HTTP_202_ACCEPTED)

    pull_security_groups_validators = [core_validators.StateValidator(models.Tenant.States.OK)]

    @decorators.detail_route(methods=['post'])
    def change_password(self, request, uuid=None):
        serializer = self.get_serializer(instance=self.get_object(), data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.TenantChangeUserPasswordExecutor.execute(self.get_object())
        return response.Response({'status': _('Password update has been scheduled.')}, status=status.HTTP_202_ACCEPTED)

    change_password_serializer_class = serializers.TenantChangePasswordSerializer
    change_password_validators = [core_validators.StateValidator(models.Tenant.States.OK)]

    @decorators.detail_route(methods=['post'])
    def pull_quotas(self, request, uuid=None):
        executors.TenantPullQuotasExecutor.execute(self.get_object())
        return response.Response({'status': _('Quotas pull has been scheduled.')}, status=status.HTTP_202_ACCEPTED)

    pull_quotas_validators = [core_validators.StateValidator(models.Tenant.States.OK)]


class NetworkViewSet(structure_views.BaseResourceViewSet):
    queryset = models.Network.objects.all()
    serializer_class = serializers.NetworkSerializer
    filter_class = filters.NetworkFilter

    disabled_actions = ['create']
    update_executor = executors.NetworkUpdateExecutor
    delete_executor = executors.NetworkDeleteExecutor
    pull_executor = executors.NetworkPullExecutor

    @decorators.detail_route(methods=['post'])
    def create_subnet(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        subnet = serializer.save()

        executors.SubNetCreateExecutor.execute(subnet)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_subnet_validators = [core_validators.StateValidator(models.Network.States.OK)]
    create_subnet_serializer_class = serializers.SubNetSerializer


class SubNetViewSet(structure_views.BaseResourceViewSet):
    queryset = models.SubNet.objects.all()
    serializer_class = serializers.SubNetSerializer
    filter_class = filters.SubNetFilter

    disabled_actions = ['create']
    update_executor = executors.SubNetUpdateExecutor
    delete_executor = executors.SubNetDeleteExecutor
    pull_executor = executors.SubNetPullExecutor



import logging

from .log import event_logger
from . import models, helpers


logger = logging.getLogger(__name__)


def log_invoice_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.paypal_invoice.info(
            '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been created.',
            event_type='invoice_creation_succeeded',
            event_context={
                'invoice': instance,
            })
    else:
        event_logger.paypal_invoice.info(
            '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been updated.',
            event_type='invoice_update_succeeded',
            event_context={
                'invoice': instance,
            })


def log_invoice_delete(sender, instance, **kwargs):
    event_logger.paypal_invoice.info(
        '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been deleted.',
        event_type='invoice_deletion_succeeded',
        event_context={
            'invoice': instance,
        })


def create_invoice(sender, invoice, issuer_details, **kwargs):
    """
    Creates an invoice when customer is "billed".
    :param sender: Invoice model
    :param invoice: Invoice instance
    :param issuer_details: details about issuer
    """
    if not invoice.items:
        return

    price = sum([item.price for item in invoice.items])

    if not price:
        return

    paypal_invoice = models.Invoice(
        customer=invoice.customer,
        year=invoice.year,
        month=invoice.month,
        invoice_date=invoice.invoice_date,
        end_date=invoice.due_date,
        tax_percent=invoice.tax_percent,
        issuer_details=issuer_details)

    paypal_invoice.payment_details = {
        'name': invoice.customer.name,
        'address': invoice.customer.address,
        'country': invoice.customer.country,
        'country_name': invoice.customer.get_country_display(),
        'email': invoice.customer.email,
        'postal': invoice.customer.postal,
        'phone_number': invoice.customer.phone_number,
        'bank_name': invoice.customer.bank_name,
        'bank_account': invoice.customer.bank_account,
    }

    paypal_invoice.save()

    for item in invoice.items:
        models.InvoiceItem.objects.create(
            invoice=paypal_invoice,
            price=item.price,
            tax=item.tax,
            quantity=helpers.get_invoice_item_quantity(item),
            unit_price=item.unit_price,
            unit_of_measure=helpers.convert_unit_of_measure(item.unit),
            name=item.name,
            start=item.start,
            end=item.end)



from waldur_ansible.jupyter_hub_management import models


def build_virtual_env_extra_args(jupyter_hub_virtual_env_request):
    return dict(
        virtual_env_name=jupyter_hub_virtual_env_request.virtual_env_name
    )


def build_delete_jupyter_hub_extra_args(delete_jupyter_hub_request):
    return dict(
        all_jupyterhub_users=map(lambda u: u.username, delete_jupyter_hub_request.jupyter_hub_management.jupyter_hub_users.all()),
        first_admin_username=delete_jupyter_hub_request.jupyter_hub_management.get_admin_users()[0].username
    )


def build_sync_config_extra_args(sync_config_request):
    def user_password_pair_builder(user):
        return dict(username=user.username, password=user.password)

    persisted_oauth_config = sync_config_request.jupyter_hub_management.jupyter_hub_oauth_config

    all_jupyter_hub_users = sync_config_request.jupyter_hub_management.jupyter_hub_users.all()
    jupyterhub_whitelisted_users = sync_config_request.jupyter_hub_management.get_whitelisted_users()
    extra_vars = dict(
        session_timeout_seconds=sync_config_request.jupyter_hub_management.session_time_to_live_hours * 3600,
        all_jupyterhub_users=map(user_password_pair_builder, all_jupyter_hub_users),
        jupyterhub_admin_users=map(user_password_pair_builder, sync_config_request.jupyter_hub_management.get_admin_users()),
        jupyterhub_whitelisted_users=map(user_password_pair_builder, jupyterhub_whitelisted_users if persisted_oauth_config else all_jupyter_hub_users),
    )

    if persisted_oauth_config:
        oauth_config = dict(
            type=persisted_oauth_config.type,
            oauth_callback_url=persisted_oauth_config.oauth_callback_url,
            client_id=persisted_oauth_config.client_id,
            client_secret=persisted_oauth_config.client_secret,
            gitlab_host=persisted_oauth_config.gitlab_host,
            tenant_id=persisted_oauth_config.tenant_id if persisted_oauth_config.type == models.JupyterHubOAuthType.AZURE else None)
        extra_vars['oauth_config'] = oauth_config

    return extra_vars



from waldur_core.logging.loggers import EventLogger, event_logger


class PriceEstimateEventLogger(EventLogger):

    class Meta:
        event_types = ('project_price_limit_updated', 'customer_price_limit_updated')
        event_groups = {
            'projects': ['project_price_limit_updated'],
            'customers': ['customer_price_limit_updated'],
        }


event_logger.register('price_estimate', PriceEstimateEventLogger)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0027_delete_floating_ip_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='securitygroup',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='start_time',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='securitygroup',
            name='backend_id',
            field=models.CharField(max_length=255, blank=True),
        ),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger


class TenantQuotaLogger(EventLogger):
    quota = 'quotas.Quota'
    tenant = 'openstack.Tenant'
    limit = float
    old_limit = float

    class Meta:
        event_types = ('openstack_tenant_quota_limit_updated',)
        event_groups = {
            'resources': event_types,
        }


event_logger.register('openstack_tenant_quota', TenantQuotaLogger)



from django.apps import AppConfig
from django.db.models import signals


class InvoiceConfig(AppConfig):
    name = 'waldur_mastermind.invoices'
    verbose_name = 'Invoices'

    def ready(self):
        from waldur_core.core import signals as core_signals
        from waldur_core.structure import models as structure_models
        from waldur_mastermind.invoices.plugins import offering_registrator
        from waldur_mastermind.invoices.plugins import openstack_registrator
        from waldur_mastermind.packages import models as packages_models
        from waldur_mastermind.support import models as support_models

        from . import handlers, models, registrators

        registrators.RegistrationManager.add_registrator(
            support_models.Offering,
            offering_registrator.OfferingItemRegistrator
        )

        registrators.RegistrationManager.add_registrator(
            packages_models.OpenStackPackage,
            openstack_registrator.OpenStackItemRegistrator
        )

        signals.post_save.connect(
            handlers.add_new_openstack_package_details_to_invoice,
            sender=packages_models.OpenStackPackage,
            dispatch_uid='waldur_mastermind.invoices.add_new_openstack_package_details_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.update_invoice_on_openstack_package_deletion,
            sender=packages_models.OpenStackPackage,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_on_openstack_package_deletion',
        )

        signals.post_save.connect(
            handlers.add_new_offering_details_to_invoice,
            sender=support_models.Offering,
            dispatch_uid='waldur_mastermind.invoices.add_new_offering_details_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.update_invoice_on_offering_deletion,
            sender=support_models.Offering,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_on_offering_deletion',
        )

        signals.pre_save.connect(
            handlers.set_tax_percent_on_invoice_creation,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.set_tax_percent_on_invoice_creation',
        )

        signals.post_save.connect(
            handlers.log_invoice_state_transition,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.log_invoice_state_transition',
        )

        signals.post_save.connect(
            handlers.emit_invoice_created_event,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.emit_invoice_created_event',
        )

        for index, model in enumerate(models.InvoiceItem.get_all_models()):
            signals.post_save.connect(
                handlers.set_project_name_on_invoice_item_creation,
                sender=model,
                dispatch_uid='waldur_mastermind.invoices.'
                             'set_project_name_on_invoice_item_creation_%s_%s' % (index, model.__class__),
            )

            signals.post_save.connect(
                handlers.update_current_cost_when_invoice_item_is_updated,
                sender=model,
                dispatch_uid='waldur_mastermind.invoices.'
                             'update_current_cost_when_invoice_item_is_updated_%s_%s' %
                             (index, model.__class__),
            )

            signals.post_delete.connect(
                handlers.update_current_cost_when_invoice_item_is_deleted,
                sender=model,
                dispatch_uid='waldur_mastermind.invoices.'
                             'update_current_cost_when_invoice_item_is_deleted_%s_%s' %
                             (index, model.__class__),
            )

        signals.post_save.connect(
            handlers.update_invoice_item_on_project_name_update,
            sender=structure_models.Project,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_item_on_project_name_update',
        )

        core_signals.pre_delete_validate.connect(
            handlers.prevent_deletion_of_customer_with_invoice,
            sender=structure_models.Customer,
            dispatch_uid='waldur_mastermind.invoices.prevent_deletion_of_customer_with_invoice',
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields
from decimal import Decimal
import django.db.models.deletion
import django.core.validators
import waldur_mastermind.invoices.utils


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0037_remove_customer_billing_backend_id'),
        ('packages', '0002_openstack_packages'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('month', models.PositiveSmallIntegerField(default=waldur_mastermind.invoices.utils.get_current_month, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)])),
                ('year', models.PositiveSmallIntegerField(default=waldur_mastermind.invoices.utils.get_current_year)),
                ('state', models.CharField(default='pending', max_length=7, choices=[('billed', 'Billed'), ('paid', 'Paid'), ('pending', 'Pending')])),
                ('customer', models.ForeignKey(verbose_name='organization', related_name='+', to='structure.Customer')),
            ],
        ),
        migrations.CreateModel(
            name='OpenStackItem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('package_details', waldur_core.core.fields.JSONField(default={}, help_text='Stores data about package', blank=True)),
                ('price', models.DecimalField(help_text='Price is calculated on a monthly basis.', max_digits=13, decimal_places=7, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_start, help_text='Date and time when item usage has started.')),
                ('end', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_end, help_text='Date and time when item usage has ended.')),
                ('invoice', models.ForeignKey(related_name='openstack_items', to='invoices.Invoice')),
                ('package', models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, to='packages.OpenStackPackage', null=True)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='invoice',
            unique_together=set([('customer', 'month', 'year')]),
        ),
    ]



from __future__ import unicode_literals

from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import admin as core_admin

from . import models, tasks


class InvoiceItemInline(core_admin.UpdateOnlyModelAdmin, admin.TabularInline):
    model = models.InvoiceItem
    readonly_fields = ('name', 'price', 'unit_price', 'unit', 'start', 'end',
                       'project_name', 'project_uuid', 'product_code', 'article_code')
    exclude = ('project',)


class OpenStackItemInline(InvoiceItemInline):
    model = models.OpenStackItem
    readonly_fields = InvoiceItemInline.readonly_fields + ('package', 'package_details')


class OfferingItemInline(InvoiceItemInline):
    model = models.OfferingItem
    readonly_fields = InvoiceItemInline.readonly_fields + ('offering', 'offering_details')


class InvoiceAdmin(core_admin.ExtraActionsMixin,
                   core_admin.UpdateOnlyModelAdmin,
                   admin.ModelAdmin):
    inlines = [OpenStackItemInline, OfferingItemInline]
    readonly_fields = ('customer', 'state', 'total', 'year', 'month')
    list_display = ('customer', 'total', 'year', 'month', 'state')
    list_filter = ('state', 'customer')
    search_fields = ('customer', 'uuid')

    def get_extra_actions(self):
        return [
            self.send_invoice_report,
            self.update_current_cost,
        ]

    def send_invoice_report(self, request):
        tasks.send_invoice_report.delay()
        message = _('Invoice report task has been scheduled')
        self.message_user(request, message)
        return redirect(reverse('admin:invoices_invoice_changelist'))

    send_invoice_report.short_description = _('Send invoice report as CSV to email')

    def update_current_cost(self, request):
        tasks.update_invoices_current_cost.delay()
        message = _('Task has been scheduled.')
        self.message_user(request, message)
        return redirect(reverse('admin:invoices_invoice_changelist'))

    send_invoice_report.short_description = _('Update current cost for invoices')


admin.site.register(models.Invoice, InvoiceAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0010_rename_floating_ip_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='backupschedule',
            name='call_count',
            field=models.PositiveSmallIntegerField(default=0, help_text='How many times backup schedule was called.'),
        ),
    ]



from django.core.management.base import BaseCommand

from ... import tasks


class Command(BaseCommand):
    help = ("Pull users from support backend.")

    def handle(self, *args, **options):
        tasks.SupportUserPullTask().run()



from __future__ import unicode_literals

from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions as rf_exceptions
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers
from waldur_core.core import signals as core_signals
from waldur_core.structure import permissions as structure_permissions
from waldur_core.structure import serializers as structure_serializers

from . import models, attribute_types


class ServiceProviderSerializer(core_serializers.AugmentedSerializerMixin,
                                serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.ServiceProvider
        fields = ('url', 'uuid', 'created', 'customer', 'customer_name', 'customer_uuid', 'description', 'enable_notifications')
        related_paths = {
            'customer': ('uuid', 'name', 'native_name', 'abbreviation')
        }
        protected_fields = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-service-provider-detail'},
            'customer': {'lookup_field': 'uuid'},
        }

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['customer'])
        return attrs


class NesterAttributeOptionSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.AttributeOption
        fields = ('key', 'title')


class NestedAttributeSerializer(serializers.ModelSerializer):
    options = NesterAttributeOptionSerializer(many=True)

    class Meta(object):
        model = models.Attribute
        fields = ('key', 'title', 'type', 'options')


class NestedSectionSerializer(serializers.ModelSerializer):
    attributes = NestedAttributeSerializer(many=True, read_only=True)

    class Meta(object):
        model = models.Section
        fields = ('key', 'title', 'attributes')


class CategorySerializer(core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    offering_count = serializers.SerializerMethodField()
    sections = NestedSectionSerializer(many=True, read_only=True)

    @staticmethod
    def eager_load(queryset):
        return queryset.prefetch_related('sections', 'sections__attributes')

    def get_offering_count(self, category):
        try:
            return category.quotas.get(name='offering_count').usage
        except ObjectDoesNotExist:
            return 0

    class Meta(object):
        model = models.Category
        fields = ('url', 'uuid', 'title', 'description', 'icon', 'offering_count', 'sections')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-category-detail'},
        }


class PlanSerializer(core_serializers.AugmentedSerializerMixin,
                     serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Plan
        fields = ('url', 'uuid', 'name', 'description', 'unit_price', 'unit', 'offering')
        protected_fields = ('offering',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
        }

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['offering'].customer)

        return attrs


class PlanWithoutOfferingSerializer(core_serializers.AugmentedSerializerMixin,
                                    serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Plan
        fields = ('url', 'uuid', 'name', 'description', 'unit_price', 'unit')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
        }


class OfferingSerializer(core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    attributes = serializers.JSONField(required=False)
    category_title = serializers.ReadOnlyField(source='category.title')
    order_item_count = serializers.SerializerMethodField()
    plans = PlanWithoutOfferingSerializer(many=True, required=False)

    class Meta(object):
        model = models.Offering
        fields = ('url', 'uuid', 'created', 'name', 'description', 'full_description', 'customer',
                  'category', 'category_title', 'rating', 'attributes', 'geolocations',
                  'is_active', 'native_name', 'native_description', 'thumbnail', 'order_item_count', 'plans')
        protected_fields = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
            'customer': {'lookup_field': 'uuid', 'view_name': 'customer-detail'},
            'category': {'lookup_field': 'uuid', 'view_name': 'marketplace-category-detail'},
        }

    def get_order_item_count(self, offering):
        try:
            return offering.quotas.get(name='order_item_count').usage
        except ObjectDoesNotExist:
            return 0

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['customer'])

        offering_attributes = attrs.get('attributes')
        if offering_attributes is not None:
            if not isinstance(offering_attributes, dict):
                raise rf_exceptions.ValidationError({
                    'attributes': 'Dictionary is expected.'
                })

            category = attrs.get('category', getattr(self.instance, 'category', None))
            self._validate_attributes(offering_attributes, category)

        return attrs

    def _validate_attributes(self, offering_attributes, category):
        offering_attribute_keys = offering_attributes.keys()
        category_attributes = list(models.Attribute.objects.filter(section__category=category,
                                                                   key__in=offering_attribute_keys))
        for key, value in offering_attributes.items():
            match_attributes = filter(lambda a: a.key == key, category_attributes)
            attribute = match_attributes[0] if match_attributes else None

            if attribute:
                klass = attribute_types.get_attribute_type(attribute.type)
                if klass:
                    try:
                        klass.validate(value, list(attribute.options.values_list('key', flat=True)))
                    except ValidationError as e:
                        err = rf_exceptions.ValidationError({'attributes': e.message})
                        raise err

    @transaction.atomic
    def create(self, validated_data):
        plans = validated_data.pop('plans', [])
        offering = super(OfferingSerializer, self).create(validated_data)
        models.Plan.objects.bulk_create([
            models.Plan(
                name=plan['name'],
                description=plan['description'],
                unit=plan['unit'],
                unit_price=plan['unit_price'],
                offering=offering,
            )
            for plan in plans
        ])
        return offering


class ScreenshotSerializer(core_serializers.AugmentedSerializerMixin,
                           serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.Screenshots
        fields = ('url', 'uuid', 'name', 'description', 'image', 'thumbnail', 'offering')
        protected_fields = ('offering', 'image')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
        }

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['offering'].customer)
        return attrs


class ItemSerializer(core_serializers.AugmentedSerializerMixin,
                     serializers.HyperlinkedModelSerializer):

    offering_name = serializers.ReadOnlyField(source='offering.name')

    class Meta(object):
        model = models.OrderItem
        fields = ('offering', 'offering_name', 'attributes', 'cost', 'plan',)
        read_only_fields = ('cost',)
        protected_fields = ('offering',)
        extra_kwargs = {
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
            'plan': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
        }

    def validate_offering(self, offering):
        if not offering.is_active:
            raise rf_exceptions.ValidationError(_('Offering is not available.'))
        return offering

    def validate(self, attrs):
        offering = attrs.get('offering')
        plan = attrs.get('plan')

        if offering and plan:
            if plan.offering != offering:
                raise rf_exceptions.ValidationError({
                    'plan': _('This plan is not available for selected offering.')
                })

        return attrs


class OrderSerializer(structure_serializers.PermissionFieldFilteringMixin,
                      core_serializers.AugmentedSerializerMixin,
                      serializers.HyperlinkedModelSerializer):

    state = serializers.ReadOnlyField(source='get_state_display')
    items = ItemSerializer(many=True)

    class Meta(object):
        model = models.Order
        fields = ('url', 'uuid',
                  'created', 'created_by', 'created_by_username', 'created_by_full_name',
                  'approved_by', 'approved_at', 'approved_by_username', 'approved_by_full_name',
                  'project', 'state', 'items', 'total_cost',)
        read_only_fields = ('created_by', 'approved_by', 'approved_at', 'state', 'total_cost')
        protected_fields = ('project', 'items')
        related_paths = {
            'created_by': ('username', 'full_name'),
            'approved_by': ('username', 'full_name'),
        }
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'created_by': {'lookup_field': 'uuid', 'view_name': 'user-detail'},
            'approved_by': {'lookup_field': 'uuid', 'view_name': 'user-detail'},
            'project': {'lookup_field': 'uuid', 'view_name': 'project-detail'},
        }

    @transaction.atomic
    def create(self, validated_data):
        user = self.context['request'].user
        validated_data['created_by'] = user
        items = validated_data.pop('items')
        order = super(OrderSerializer, self).create(validated_data)
        models.OrderItem.objects.bulk_create([
            models.OrderItem(
                order=order,
                offering=item['offering'],
                attributes=item.get('attributes', {}),
            )
            for item in items
        ])
        return order

    def get_filtered_field_names(self):
        return 'project',


def get_is_service_provider(serializer, scope):
    customer = structure_permissions._get_customer(scope)
    return models.ServiceProvider.objects.filter(customer=customer).exists()


def add_service_provider(sender, fields, **kwargs):
    fields['is_service_provider'] = serializers.SerializerMethodField()
    setattr(sender, 'get_is_service_provider', get_is_service_provider)


core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerSerializer,
    receiver=add_service_provider,
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-10 08:44
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='IdentityProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField(db_index=True)),
                ('url', models.URLField()),
                ('metadata', waldur_core.core.fields.JSONField(default={}))
            ],
            options={
                'ordering': ('name',),
            },
        ),
    ]



from __future__ import unicode_literals

import six
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import JSONField as BetterJSONField
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from django_fsm import transition, FSMIntegerField

from waldur_core.core import models as core_models
from waldur_core.core.fields import JSONField
from waldur_core.core.validators import FileTypeValidator
from waldur_core.quotas import fields as quotas_fields
from waldur_core.quotas import models as quotas_models
from waldur_core.structure import models as structure_models
from waldur_core.structure.images import get_upload_path

from . import managers
from .attribute_types import ATTRIBUTE_TYPES
from .plugins import manager
from ..common import mixins as common_mixins


@python_2_unicode_compatible
class ServiceProvider(core_models.UuidMixin,
                      core_models.DescribableMixin,
                      structure_models.StructureModel,
                      structure_models.TimeStampedModel):
    customer = models.OneToOneField(structure_models.Customer, related_name='+', on_delete=models.CASCADE)
    enable_notifications = models.BooleanField(default=True)

    class Permissions(object):
        customer_path = 'customer'

    class Meta(object):
        verbose_name = _('Service provider')

    def __str__(self):
        return six.text_type(self.customer)

    @classmethod
    def get_url_name(cls):
        return 'marketplace-service-provider'


VectorizedImageValidator = FileTypeValidator(
    allowed_types=[
        'image/png',
        'image/jpeg',
        'image/svg+xml',
    ]
)


@python_2_unicode_compatible
class Category(core_models.UuidMixin,
               quotas_models.QuotaModelMixin,
               structure_models.TimeStampedModel):
    title = models.CharField(blank=False, max_length=255)
    icon = models.FileField(upload_to='marketplace_category_icons',
                            blank=True,
                            null=True,
                            validators=[VectorizedImageValidator])
    description = models.TextField(blank=True)

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        offering_count = quotas_fields.CounterQuotaField(
            target_models=lambda: [Offering],
            path_to_scope='category',
        )

    class Meta(object):
        verbose_name = _('Category')
        verbose_name_plural = _('Categories')
        ordering = ('title',)

    def __str__(self):
        return six.text_type(self.title)

    @classmethod
    def get_url_name(cls):
        return 'marketplace-category'


@python_2_unicode_compatible
class Section(structure_models.TimeStampedModel):
    key = models.CharField(primary_key=True, max_length=255)
    title = models.CharField(blank=False, max_length=255)
    category = models.ForeignKey(Category, related_name='sections')

    def __str__(self):
        return six.text_type(self.title)


@python_2_unicode_compatible
class Attribute(structure_models.TimeStampedModel):
    key = models.CharField(primary_key=True, max_length=255)
    title = models.CharField(blank=False, max_length=255)
    section = models.ForeignKey(Section, related_name='attributes')
    type = models.CharField(max_length=255, choices=ATTRIBUTE_TYPES)

    def __str__(self):
        return six.text_type(self.title)


@python_2_unicode_compatible
class AttributeOption(models.Model):
    attribute = models.ForeignKey(Attribute, related_name='options', on_delete=models.CASCADE)
    key = models.CharField(max_length=255)
    title = models.CharField(max_length=255)

    class Meta(object):
        unique_together = ('attribute', 'key')

    def __str__(self):
        return six.text_type(self.title)


@python_2_unicode_compatible
class Offering(core_models.UuidMixin,
               core_models.NameMixin,
               core_models.DescribableMixin,
               quotas_models.QuotaModelMixin,
               structure_models.StructureModel,
               structure_models.TimeStampedModel):
    thumbnail = models.ImageField(upload_to='marketplace_service_offering_thumbnails', blank=True, null=True)
    full_description = models.TextField(blank=True)
    rating = models.IntegerField(null=True,
                                 validators=[MaxValueValidator(5), MinValueValidator(1)],
                                 help_text=_('Rating is value from 1 to 5.'))
    category = models.ForeignKey(Category, related_name='offerings')
    customer = models.ForeignKey(structure_models.Customer, related_name='+', null=True)
    attributes = BetterJSONField(blank=True, default=dict)
    geolocations = JSONField(default=list, blank=True,
                             help_text=_('List of latitudes and longitudes. For example: '
                                         '[{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]'))
    is_active = models.BooleanField(default=True)

    native_name = models.CharField(max_length=160, default='', blank=True)
    native_description = models.CharField(max_length=500, default='', blank=True)

    content_type = models.ForeignKey(ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')

    objects = managers.OfferingManager()

    class Permissions(object):
        customer_path = 'customer'

    class Meta(object):
        verbose_name = _('Offering')

    class Quotas(quotas_models.QuotaModelMixin.Quotas):
        order_item_count = quotas_fields.CounterQuotaField(
            target_models=lambda: [OrderItem],
            path_to_scope='offering',
        )

    def __str__(self):
        return six.text_type(self.name)

    @classmethod
    def get_url_name(cls):
        return 'marketplace-offering'


@python_2_unicode_compatible
class Screenshots(core_models.UuidMixin,
                  structure_models.StructureModel,
                  core_models.DescribableMixin,
                  structure_models.TimeStampedModel,
                  core_models.NameMixin):
    image = models.ImageField(upload_to=get_upload_path)
    thumbnail = models.ImageField(upload_to=get_upload_path, editable=False, null=True)
    offering = models.ForeignKey(Offering, related_name='screenshots')

    class Permissions(object):
        customer_path = 'offering__customer'

    class Meta(object):
        verbose_name = _('Screenshot')

    def __str__(self):
        return six.text_type(self.name)

    @classmethod
    def get_url_name(cls):
        return 'marketplace-screenshot'


class Order(core_models.UuidMixin,
            structure_models.TimeStampedModel):
    class States(object):
        DRAFT = 1
        REQUESTED_FOR_APPROVAL = 2
        EXECUTING = 3
        DONE = 4
        TERMINATED = 5

        CHOICES = (
            (DRAFT, 'draft'),
            (REQUESTED_FOR_APPROVAL, 'requested for approval'),
            (EXECUTING, 'executing'),
            (DONE, 'done'),
            (TERMINATED, 'terminated'),
        )

    created_by = models.ForeignKey(core_models.User, related_name='orders')
    approved_by = models.ForeignKey(core_models.User, blank=True, null=True, related_name='+')
    approved_at = models.DateTimeField(editable=False, null=True, blank=True)
    project = models.ForeignKey(structure_models.Project)
    state = FSMIntegerField(default=States.DRAFT, choices=States.CHOICES)
    total_cost = models.DecimalField(max_digits=22, decimal_places=10, null=True, blank=True)

    class Permissions(object):
        customer_path = 'project__customer'
        project_path = 'project'

    class Meta(object):
        verbose_name = _('Order')
        ordering = ('created',)

    @classmethod
    def get_url_name(cls):
        return 'marketplace-order'

    @transition(field=state, source=States.DRAFT, target=States.REQUESTED_FOR_APPROVAL)
    def set_state_requested_for_approval(self):
        pass

    @transition(field=state, source=States.REQUESTED_FOR_APPROVAL, target=States.EXECUTING)
    def set_state_executing(self):
        pass

    @transition(field=state, source=States.EXECUTING, target=States.DONE)
    def set_state_done(self):
        pass

    @transition(field=state, source='*', target=States.TERMINATED)
    def set_state_terminated(self):
        pass


class OrderItem(core_models.UuidMixin,
                structure_models.TimeStampedModel):
    order = models.ForeignKey(Order, related_name='items')
    offering = models.ForeignKey(Offering)
    attributes = BetterJSONField(blank=True, default=dict)
    cost = models.DecimalField(max_digits=22, decimal_places=10, null=True, blank=True)
    plan = models.ForeignKey('Plan', null=True, blank=True)

    content_type = models.ForeignKey(ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')
    objects = managers.OrderItemManager('scope')

    class Meta(object):
        verbose_name = _('Order item')
        ordering = ('created',)

    def process(self, user):
        manager.process(self, user)


class Plan(core_models.UuidMixin,
           structure_models.TimeStampedModel,
           core_models.NameMixin,
           core_models.DescribableMixin,
           common_mixins.UnitPriceMixin):
    offering = models.ForeignKey(Offering, related_name='plans')

    @classmethod
    def get_url_name(cls):
        return 'marketplace-plan'

    class Permissions(object):
        customer_path = 'offering__customer'



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class AuthSocialExtension(WaldurExtension):

    class Settings:
        # wiki: https://opennode.atlassian.net/wiki/display/WD/AuthSocial+plugin+configuration
        WALDUR_AUTH_SOCIAL = {
            'GOOGLE_SECRET': 'PLACEHOLDER',
            'GOOGLE_CLIENT_ID': 'PLACEHOLDER',

            'FACEBOOK_SECRET': 'PLACEHOLDER',
            'FACEBOOK_CLIENT_ID': 'PLACEHOLDER',

            'SMARTIDEE_SECRET': 'PLACEHOLDER',
            'SMARTIDEE_CLIENT_ID': 'PLACEHOLDER',

            'USER_ACTIVATION_URL_TEMPLATE': 'http://example.com/#/activate/{user_uuid}/{token}/',
        }

    @staticmethod
    def get_public_settings():
        return ['GOOGLE_CLIENT_ID', 'FACEBOOK_CLIENT_ID', 'SMARTIDEE_CLIENT_ID']

    @staticmethod
    def django_app():
        return 'waldur_auth_social'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns
